//  Copyright (c) 1997-2004 by Mark F. Adams 
//  $Id: agg_smooth.C,v 1.51 2005/09/15 17:27:30 adams Exp $
//  Author:             Mark F. Adams
//  Filename:           amg_smooth.C
//
#include "prometheus.hh"
  
#undef __FUNCT__
#define __FUNCT__ "PromAggGrid::SmoothP" 
/* PromAggGrid::SmoothP *******************************************************
 *
 *  - replace next->restriction: P := P - wD^-1AP 
 *  - collective on 'wcomm'
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *
 *  RETURN: 
 *     - PETSc error code 
 */ 
int PromAggGrid::SmoothP() const
{ 
  int    ierr,pp,np; assert(stiffness_ != NULL); double l1;
  const PromMatrix * const AA = stiffness_;  assert(AA!=NULL);
  
  perf_mon_.EventBegin( PromContext::prom_logs_[LOG_MAPS] );

  // lambda est.
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS0]);  
#if !defined(PROM_NO_STAGES)
  perf_mon_.StagePush(PromContext::prom_stages_[5]);
#endif
  // collective on 'wcomm'
  ierr = AA->ComputeLambda( options_, perf_mon_, 20 ); CHKERRQ(ierr);
  l1 = AA->getDRho();
#if !defined(PROM_NO_STAGES)
  perf_mon_.StagePop();
#endif
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS0]);

  if( isActive() ) {
    // DAP <- AP
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS1]);
    ierr = FormDAP( AA );	                 CHKERRQ(ierr);

    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS1]);

    // DAP <- (I - mD^-1A)P
    ierr = ApplySmoothing( AA, 1.5 / (l1*1.1) );           CHKERRQ(ierr);

    // output P
    if( comm_.np() == -1 ){ 
      char name[25];
      sprintf( name, "P_1_level_%d", getLevel() );
      FILE *file = fopen( name, "w" ); 
      for( int xx = 0 ; xx < nNodes_ ; xx++ ){
	PromAggNode *fine = dynamic_cast<PromAggNode*>(nodes_[xx]);
	int geqi, geqj, fndf = fine->getNDF(), ii, jj; 
	const PromParentArr *parI = fine->parents_; 
	for( pp = 0, np = fine->nParents() ; pp < np ; pp++ ){
	  int cndf = parI->data_[pp]->getNDF();
	  assert(parI->data_[pp]->DAP_ != NULL); // no!!
	  for(ii = 0, geqi = fine->geq() ; ii < fndf ; ii++, geqi++){
	    const shpfloat * const pp0 = &parI->data_[pp]->DAP_[ii*cndf]; 
	    for(jj=0, geqj=parI->data_[pp]->geq() ; jj<cndf ; jj++, geqj++){
	      if(fabs(pp0[jj]) > 1.e-12 )
		fprintf(file,"%7d%7d%15.7e\n",geqi,geqj,pp0[jj]);
	    }
	  }
	}
      }
      fclose(file);
    }
  }
  else {
    perf_mon_.EventBegin( PromContext::prom_logs_[LOG_MIS1] );
    perf_mon_.EventEnd( PromContext::prom_logs_[LOG_MIS1] );
  }
  perf_mon_.Flops(1); // make sure this is not in symbolic
  perf_mon_.EventEnd( PromContext::prom_logs_[LOG_MAPS] );

  return 0;
}
    
#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::FormDAP" 
/* PromAggGrid::FormDAP ************************************************
 *
 *  - form D^-1 A P: "parents->DAP" <- D^-1 A "parents->P0_"  
 *
 *  INPUT:
 *   - AA: stiffness matrix
 *
 *  SIDE EFFECTS:
 *   - allocs DAP
 *
 *  RETURN: 
 *     - PETSc error code 
 */ 
int PromAggGrid::FormDAP( const PromMatrix * const AA ) const
{
  const int  mype=comm_.mype(),my0=my0Nd(); PromParent *pari;
  shpfloat *pDAP; const int fact=AA->isComplex() ? 2 : 1,fact2=fact*fact;
  int        ierr,ii,xx,ncols,pp,flops=0,adjacs[PROM_MAXNIEGH];
  double    *Ai[PROM_MAXNIEGH];  assert(AA != NULL);
  if(AA->isSymmetricMatrix())SETERRQ(-1,"FormDAP called with symmetic matrix");

  for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
    PromNode *finei = nodes_[xx];
    if( finei->skipSmooth() ) continue; // skip nodes that do not get smoothed
    PromParentArr *parIarr = finei->parents_; 
    PromParent **parIdata = parIarr->data_; 
    int nparI = finei->nParents();
    int fndfj, fndfi = finei->getNDF();
    const int nparOr = nparI;

    if( fndfi == 4 ) { // hack for accoustic/structure
      ierr = finei->SetSkipSmooth();  CHKERRQ(ierr);
      assert(0); continue;
    }

    // row of fine matrix
    ierr = AA->GetLocalNodeRow(xx, PROM_MAXNIEGH, ncols, adjacs, Ai); 
    CHKERRQ(ierr); // just use real part of 'Ai'
    // for whole row, get parent array and ndf
    for( int col = 0; col < ncols ; col++ ) {
      const int gidj=adjacs[col],lidj=gidj-my0,isloc=(lidj>=0&&lidj<nLocalNd_);
      PromNode *finej =  isloc ? nodes_[lidj] : ghosts_.find(gidj+1);
      const PromParentArr *parJarr = NULL;
      if( finej != NULL ) {
	fndfj = finej->getNDF(); parJarr = finej->parents_;
      }
      else {
	PromParentArr_extra *gpararr = gid_extraParents_->find(gidj+1);
	assert(gpararr!=NULL); assert(gpararr->getGID() == gidj);
	parJarr = gpararr; fndfj = gpararr->getNDF();
      }
      const double *const Aij = Ai[col]; 

      // apply DAP += A_ij * P_j
      const PromParent * const parj0 = parJarr->data_[0];
      const int pargidJ = parj0->global(), cndfj = parj0->getNDF(); assert(cndfj>0);
      // find pari 
      for( pp = 0, pari = NULL ; pp < nparI && pari == NULL; pp++ ){
	if(parIdata[pp]->global() == pargidJ) pari = parIdata[pp];
      }
      // make extra parent caused by regularization/penalty or reduction of top
      if( pari == NULL ) {
	assert(isTop());
	ierr = finei->SetnParents( parBank_, nparI+1 );  CHKERRQ(ierr);
	parIarr = finei->parents_;  assert(gid_extraParents_!=NULL);
	parIdata = parIarr->data_; 
	pari = parIdata[nparI++];
	pari->SetNDF( parj0->getNDF() );
	pari->setGlobal( pargidJ );
	pari->setProc( parj0->proc() );
	pari->setGEQ( parj0->geq() );
      }

      // alloc and zero if necc.
      if( (pDAP=pari->DAP_) == NULL ) {
	int sz = cndfj*fndfi, bsz = sizeof(shpfloat)*sz; assert(bsz>0);
	ierr = PetscMalloc( bsz, &pDAP );  CHKERRQ(ierr); 
	pari->DAP_ = pDAP; PetscMemzero( pDAP, bsz ); 
	//pDAP[sz] = PROM_BIG_NUM;
	if( pari->getPsz() == 0 ) {
	  ierr = pari->SetPsz_hard( sz ); CHKERRQ(ierr);
	}
	else if(pari->getPsz() != sz){
	  SETERRQ4(1,"locid=%d, cndfj=%d par->psz()=%d, sz=%d",xx,cndfj,pari->getPsz(),sz);
	}
      }
      else if(pari->getPsz() != cndfj*fndfi) SETERRQ4(1,"locid=%d, cndfj=%d par->psz()=%d, sz=%d",xx,cndfj,pari->getPsz(),cndfj*fndfi);
      //assert(pDAP[cndfj*fndfi] == PROM_BIG_NUM);

      // do DAP += A_ij * P_j
      if( cndfj == 1 ) *pDAP += parj0->getScalarP0() * *Aij;
      else{
	const shpfloat* const P0 = parj0->P0_; assert(P0 != NULL);
	const double *aijT = Aij; // transpose
	int iii,ij,jj,jjj,ii;
	// buffer := A_ij^T * P_jJ 
	if( cndfj == 6 && fndfi%3 == 0 ) { // unroll
	  for( ii = 0, iii = 0, ij = 0 ; ii < fndfj ; ii++, iii += 6 ){
	    shpfloat t0 = P0[iii],  t1 = P0[iii+1],t2 = P0[iii+2];
	    shpfloat t3 = P0[iii+3],t4 = P0[iii+4],t5 = P0[iii+5];
	    for(jj = jjj = 0 ; jj < fndfi ; jj += 3, ij += 3*fact2, jjj += 18){
	      shpfloat *pv = &pDAP[jjj];
	      double aij0=aijT[ij],aij1=aijT[ij+fact],aij2=aijT[ij+2*fact];
	      pv[0] += aij0 * t0;	    pv[1] += aij0 * t1;
	      pv[2] += aij0 * t2;	    pv[3] += aij0 * t3;
	      pv[4] += aij0 * t4;	    pv[5] += aij0 * t5;
	      pv[6] += aij1 * t0;	    pv[7] += aij1 * t1;
	      pv[8] += aij1 * t2;	    pv[9] += aij1 * t3;
	      pv[10] += aij1 * t4;	    pv[11] += aij1 * t5;
	      pv[12] += aij2 * t0;	    pv[13] += aij2 * t1;
	      pv[14] += aij2 * t2;	    pv[15] += aij2 * t3;
	      pv[16] += aij2 * t4;	    pv[17] += aij2 * t5;
	    }
	  }
	}
	else {
	  const double *paijt = &aijT[0]; shpfloat *pv = &pDAP[0];
	  for( ii = 0 ; ii < fndfi ; ii++, paijt++ ) {
	    const shpfloat * pp0 = &P0[0], *p0;  
	    for( int jj = 0, kk ; jj < cndfj ; jj++, pv++, pp0++ ) {
	      const double *pa = paijt + ii*fact; p0 = pp0; 
	      for(kk=0;kk<fndfj;kk++,p0+=cndfj,pa+=fndfi*fact2) *pv += *pa * *p0;
	    }
	  }
	}
      }
      flops += fndfj*fndfi*cndfj;
      //assert(pDAP[cndfj*fndfi] == PROM_BIG_NUM);
    }
    // skip smoothing to keep npar <  PROM_MAXPAR_SOFT -- invarient!!!
    if( ((nparI > 2*nparOr && nparI > PROM_MAXPAR_SOFT/2) || 
	 nparI > PROM_MAXPAR_SOFT ) ) { //
      for( int pp = 0 ; pp < nparI ; pp++ ){
	PromParent *par = parIdata[pp];
	if( par->DAP_ != NULL ) PetscFree( par->DAP_ ); 
	par->DAP_ = NULL; 
      }
      ierr = finei->SetnParents( parBank_, 1 );  CHKERRQ(ierr);
      ierr = finei->SetSkipSmooth();  CHKERRQ(ierr);
    }
  }
  perf_mon_.Flops(2*flops); flops = 0;
#if defined(PETSC_USE_BOPT_g)
  ierr = PetscTrValid(__LINE__, __FUNCT__, __FILE__, __SDIR__); CHKERRQ(ierr);
#endif

  // v <- D^-1 v 
#if defined(PROM_USE_NODAL_SMAGG)
  int zzs=0;
  double Ainv[4*PROM_MAXNDF*PROM_MAXNDF],tarr[4*PROM_MAXNDF*PROM_MAXNDF];
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
    PromNode *finei = nodes_[xx];
    if( finei->skipSmooth() ) continue; // skip nodes that do not get smoothed
    ierr = AA->GetLocalNodeRow(xx,PROM_MAXNIEGH,ncols,adjacs,Ai);CHKERRQ(ierr);
    int ndf=finei->getNDF(),ndf2=ndf*ndf,igid=finei->index_.global(),done,jj;
    for( jj = 0, done = 0 ; jj < ncols ; jj++ ) {
      int jgid = adjacs[jj];
      if( igid == jgid ) {
	const double * const aijT = Ai[jj];
	zzs += PromContext::TInvert(aijT,Ainv,ndf,ndf*fact);
	done++; 
	break;
      }
    } 
    assert(done==1);
    // smooth
    for( int pp = 0 ; pp < finei->nParents() ; pp++ ) {
      PromParent *par = finei->parents_->data_[pp];
      if( par->DAP_ != NULL ) {
	assert(par->getPsz()%ndf == 0);
	const int cndf = par->getPsz()/ndf; assert(cndf>0);
	shpfloat *pB, *pv = Ainv;           assert(cndf==par->getNDF());
	for( ii = 0 ; ii < ndf ; ii++, pv += ndf ) {
	  for( jj = 0, pB = par->DAP_ ; jj < cndf ; jj++, pB++ ) {
	    double t = 0.0; 
	    for( int kk = 0 ; kk < ndf ; kk++ ) t += pv[kk] * pB[cndf*kk];
	    tarr[cndf*ii + jj] = t;
	  }
	}
	for( ii = 0, pv = tarr, pB = par->DAP_ ; ii < ndf ; ii++ ){
	  for(int jj = 0 ; jj < cndf ; jj++ ) *pB++ = *pv++;
	}
	flops += cndf*ndf2 + ndf2*ndf;
      }
    }
  }
  if( zzs ) {
    PetscPrintf(PETSC_COMM_SELF,
		"[%d] *WARNING* %d Zero pivot(s) in INVERT N=%d\n",
		mype,zzs,AA->getM() );
  }
  perf_mon_.Flops(2*flops); flops = 0;
#else
  double *diag; int nneq=AA->map_.nlocalEq(),eqi,leq,iszero;
  PromVector *work = new PromVector( *this );
  ierr = AA->GetDiagonal( work );                     CHKERRQ(ierr);
  ierr = work->GetArray( &diag );                         CHKERRQ(ierr);
  for(xx=ncols=leq=0,eqi=proc_geq_[mype] ; xx < nLocalNd_ ; xx++){
    PromNode *finei = nodes_[xx]; 
    if( finei->skipSmooth() ) continue; // skip nodes that do not get smoothed
    const int fndf = finei->getNDF();
    for( int pp = 0 ; pp < finei->nParents() ; pp++ ){
      PromParent *par = finei->parents_->data_[pp];
      const int cndf = par->getNDF(); assert(cndf>0);
      shpfloat *pDAP = par->DAP_;
      if( pDAP != NULL ) {
	for(int ii = 0, jj = leq, tt = 0 ; ii < fndf ; ii++, jj++ ){
	  if( pp==0 && diag[jj] < 1.e-100 ) {
	    PetscPrintf(PETSC_COMM_SELF,
			"\t[%d]%s warning diag[%d,ndf %d]=%e\n",
			mype,__FUNCT__,xx,ii,diag[jj]);
	    ncols++;   diag[jj] = 1.0;   iszero = 1;
	  }
	  else iszero = 0;
	  if( finei->isFixed(ii) || iszero == 1 ){
	    for(int kk = 0 ; kk < cndf ; kk++, tt++ ) pDAP[tt] = 0.0;
	  }
	  else {
	    shpfloat dtemp = 1. / diag[jj];
	    for(int kk=0 ; kk<cndf ; kk++, tt++ ) pDAP[tt] *= dtemp;
	  }
	}
	flops += cndf*fndf;
      }
    }
    eqi += fndf;
    leq += fndf;
  }
  ierr = work->RestoreArray( &diag );                     CHKERRQ(ierr);
  delete work;
  perf_mon_.Flops(2*flops+nneq); flops=0;
#endif

  return 0;
} 

#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::checkParents" 
/* PromAggGrid::checkParents *********************************************
 *
 *  RETURN:
 *     - PETSc error code 
 */
int PromAggGrid::checkParents()const
{
  if( isActive() ){
    int ii,xx,ierr,*PType,nLoc=nlocal(),*lnPar,mype=comm_.mype(),np=comm_.np();
    ierr = PetscMalloc(nglobal()*sizeof(int), &PType); CHKERRQ(ierr);
    ierr = PetscMalloc((nLoc+1)*sizeof(int), &lnPar); CHKERRQ(ierr);
    PetscPrintf(gComm_,"\t\t[%d]%s start ng=%d\n",mype,__FUNCT__,nglobal()); 
    for(xx=0;xx<nNodes_;xx++){
      PromNode *finei = nodes_[xx]; 
      const int nparI = finei->nParents();
      for( int pp = 0 ; pp < nparI ; pp++ ){
	PromParent *par = finei->parents_->data_[pp];
	if(par->getNDF()<=0)
	  SETERRQ2(1,"par->ndf()=%d, isloc=%d",par->getNDF(),xx<nLoc);
	assert(par->global() != 0 || par->proc() == 0 );
      }
      if(xx<nLoc)lnPar[xx]=nparI;
    }
    // 
    PetscPrintf(gComm_,"\t\t[%d]%s Alltoall nl=%d\n",mype,__FUNCT__,nLoc);
    int recvcounts[PROM_BSZ],proc_gnode[PROM_BSZ];
    for( xx = ii = 0 ; xx < np ; xx += factor_, ii++ ){
      recvcounts[ii] = proc_gnode_[xx+1] - proc_gnode_[xx];
      proc_gnode[ii] = proc_gnode_[xx];
    }
    MPI_Allgatherv(lnPar,nLoc,MPI_INT,PType,recvcounts,proc_gnode,MPI_INT,
		   gComm_); 
    PetscPrintf(gComm_,"\t\t[%d]%s Alltoall done\n",mype,__FUNCT__);
    for(xx=nLoc;xx<nNodes_;xx++){
      PromNode *ghos = nodes_[xx]; 
      if(PType[ghos->index_.global()]!=ghos->nParents() && ghos->nParents()!=1)
	SETERRQ3(1,"ghost %d, npar=%d, real npar=%d",ghos->index_.global(),
		 ghos->nParents(),PType[ghos->index_.global()]);
    }
    PetscFree(lnPar); PetscFree(PType);
    PetscPrintf(gComm_,"\t\t[%d]%s done\n",mype,__FUNCT__); 
  }

  return 0;
}

//#define SKIP_NEAR_BOUND
#undef __FUNCT__  
#define __FUNCT__ "PromAggGrid::ApplySmoothing" 
/* PromAggGrid::ApplySmoothing *********************************************
 * 
 *  - Put P1 into DAP: DAP := P - omega DAP
 *
 *  INPUT:
 *   - AA:
 *   - omega:
 *
 *  SIDE EFFECTS:
 *   - collect DAP partials from ghosts
 *
 *  RETURN:
 *     - PETSc error code 
 */
int PromAggGrid::ApplySmoothing(const PromMatrix*const AA, double omega ) const
{
  int  ierr,jj,tt,ii,lev=getLevel(),flops=0,proc,ncols,todo,mype=comm_.mype();
  int  xx,adjacs[PROM_MAXNIEGH],my0=my0Nd();
  shpfloat *pDAP, *pP0; bool verb = (options_.verbose_>1 && mype==0);
  xx=sizeof(shpfloat); jj=sizeof(MPI_Request);  ii=sizeof(int);
  const int fisz = xx/ii, reqsz = (jj/xx + ((jj%xx == 0) ? 0 : 1)) * fisz;
  MPI_Status  status; 
  PromList<PromNode*> *list;
  PromList<MPI_Request*> smessages;
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromList<int> procs(10); const int tag = getNewTag(); 

  // mark 'active' (skip smoothing for 'inactive'), collect send lists
  for( xx = todo = 0 ; xx < nLocalNd_ ; xx++ ){
    PromNode *finei = nodes_[xx]; 
    const int nparI = finei->nParents();
#if defined(SKIP_NEAR_BOUND)
    int isfx = 0;
#endif
    ierr = AA->GetLocalNodeRow(xx,PROM_MAXNIEGH,ncols,adjacs);CHKERRQ(ierr);
    for( jj = 0 ; jj < ncols ; jj++ ) {
      const int gidj=adjacs[jj],lid=gidj-my0,isloc=(lid>=0 && lid<nLocalNd_);
      // check for fixed
      PromNode *finej =  isloc ? nodes_[lid] : ghosts_.find(gidj+1);
      PromParentArr_extra *gpararr = NULL;
      if( finej == NULL ) { // extra parents from regularization/penalty
	assert(gid_extraParents_!=NULL);
	gpararr = gid_extraParents_->find(gidj+1); assert(gpararr!=NULL);
#if defined(SKIP_NEAR_BOUND)
	if( gpararr->isFixed() ) isfx++;
#endif 
      }
      else {
#if defined(SKIP_NEAR_BOUND)
	if( finej->isFixed() ) isfx++;
#endif
      }
      // check 'todo'
      if( !isloc ) {
	if( gpararr == NULL ) {
	  if( !finej->isMarked() ) { todo++; finej->mark(); }
	}
	else if( !gpararr->isMarked() ) { todo++; gpararr->mark(); }
	// update procs
	ierr = GetProc( gidj, proc );   CHKERRQ(ierr);
	if( !procs.find(proc+1) )procs.addTail(proc+1);// tiny linear search
      }
    }

    // mark !local near bound. local mark -> active
    if( !finei->skipSmooth() ) {
#if defined(SKIP_NEAR_BOUND)
      if( isfx || finei->isFixed() ) 
#else
      if( finei->isFixed() ) 
#endif
	{
	  assert(lev==0);
	  // skip smoothing 
	  for( int pp = 0 ; pp < nparI ; pp++ ){
	    PromParent *par = finei->parents_->data_[pp];
	    if(par->DAP_!=NULL) PetscFree( par->DAP_ ); 
	    par->DAP_ = NULL;
	  }
	  ierr = finei->SetnParents( parBank_, 1 );  CHKERRQ(ierr);
	  assert(finei->parents_->data_[0]->P0_!=NULL); 
	  ierr = finei->SetSkipSmooth();       CHKERRQ(ierr);
	}
    }

    // collect bound to send
    while( !procs.isEmpty() ) {
      proc = procs.removeHead() - 1;  
      if( (list=proc_list.find(proc+1)) == NULL ) {
	list = new PromList<PromNode*> (128); CHKNEW(list,1);
	proc_list.add( list, proc+1 ); 
      }
      list->addTail( finei );
    }
  }

  // clear up unmarked ghost's pars (not touched but covered by local agg.)
  for( xx = nLocalNd_ ; xx < nNodes_ ; xx++ ){
    PromNode *finej = nodes_[xx];
    if( !finej->isMarked() ) {
      if( finej->nParents() > 1 ) {
	ierr = finej->SetnParents( parBank_, 1 );  CHKERRQ(ierr);
      }
      assert(finej->parents_->data_[0]->DAP_ == NULL);
    }
  }
  if( gid_extraParents_ != NULL ) {
    TablePos tpos = gid_extraParents_->getHeadPosition(); 
    while( tpos ) { 
      PromParentArr_extra *parr = gid_extraParents_->getNext( tpos );
      if( !parr->isMarked() ) {
	if( parr->getSize() > 1 ) {
	  ierr = parr->SetSize( parBank_, 1 );  CHKERRQ(ierr);
	}
	assert(parr->data_[0]->DAP_ == NULL);
      }
    }
  }

  // DAP := P - omega * DAP
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
    PromNode *finei = nodes_[xx]; 
    if( finei->skipSmooth() ) assert(finei->parents_->data_[0]->DAP_==NULL);
    else{
      for( jj = 0 ; jj < finei->nParents() ; jj++ ) {
	PromParent *par = finei->parents_->data_[jj]; 
	const int sz = par->getPsz(); assert(jj>0 || sz>0); 
	flops += sz; 
	if( jj == 0 && par->DAP_ != NULL ) { //  P0 != 0.0
	  if( sz == 1 ) {
	    par->DAP_[0] = par->getScalarP0() - omega * par->DAP_[0];
	  }
	  else {
	    pDAP = par->DAP_; pP0 = par->P0_;
	    for(ii=0;ii<sz;ii++) pDAP[ii] = pP0[ii] - omega * pDAP[ii];
	  }
	}
	else { // P0 == 0.0
	  if( (pDAP=par->DAP_) != NULL ) {
	    for(ii = 0 ; ii < sz ; ii++) pDAP[ii] = -omega * pDAP[ii]; // P0==0.
	  }
	}
      }
    }
  }
  perf_mon_.Flops(flops*2); flops = 0;
  if(verb)PetscPrintf(PETSC_COMM_SELF,"\t[%d]%s omega=%e\n",mype,__FUNCT__,omega);

  // send back parents
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    Position pos = list->getHeadPosition(); assert(pos!=NULL); tt = 0;
    while( pos ) {
      PromNode *finei = list->getNext( pos ); 
      const int fndf = finei->getNDF();
      tt += 2; // gid, npar
      for( jj = 0 ; jj < finei->nParents() ; jj++ ) {
	PromParent *par = finei->parents_->data_[jj];
	if(par->DAP_ != NULL) tt += par->getNDF()*fndf*fisz + 2; // psz, gidJ, ...
	else tt += 2;
      }
    }
    MPI_Request *sreq;
    ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );     
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      PromNode *finei = list->removeHead(); 
      const int fndf = finei->getNDF();
      *pb++ = finei->index_.global(); int *npar = pb++, pp;
      for( jj = 0, pp = 0 ; jj < finei->nParents() ; jj++ ) {
	PromParent *par = finei->parents_->data_[jj];
	pDAP = par->DAP_; 
	int sz = (pDAP != NULL) ? par->getNDF()*fndf : 0; 
	*pb++ = sz; *pb++ = par->global();
	shpfloat *fpb = (shpfloat*)pb;
	for( ii = 0 ; ii < sz ; ii++ ) *fpb++ = *pDAP++;
	pb = (int*)fpb;
	pp++; 
      }
      *npar = pp;
    }
    assert(tt == pb-buff);
    // send(tag) *< gid, npar, npar*< psz, gidJ, DAP[]> >
    MPI_Isend( buff, pb-buff, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list; 
  }

  // recieve parents
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt );
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag) < gidj, npar, npar*<psz, jidJ, DAP[]> >
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status );
    while( pb-buff < tt ) {
      const int gidj = *pb++, npRev = *pb++;
      PromNode *finej = ghosts_.find(gidj+1); 
      PromParentArr *pararr = NULL;  int fndf;
      if( finej != NULL ){
	assert(finej->isMarked());
	pararr = finej->parents_;  assert(pararr!=NULL); finej->mark(FALSE);
	fndf = finej->getNDF(); 
      }
      else {
	assert(gid_extraParents_!=NULL);
	PromParentArr_extra *gpararr = gid_extraParents_->find(gidj+1);
	assert( gpararr != NULL );
	pararr = gpararr; assert( gpararr != NULL );
	assert(gpararr->isMarked()); gpararr->mark( FALSE );
	fndf = gpararr->getNDF();
      }
      todo--;  //assert(todo>=0);
      const int parSzOrig = pararr->getSize();
      if( npRev != parSzOrig ){
	ierr = pararr->SetSize( parBank_, npRev );  CHKERRQ(ierr);
      }
      for( jj = 0 ; jj < npRev ; jj++ ) {
	PromParent *par = pararr->data_[jj];
	const int sz = *pb++, gidJ = *pb++; 
	if( par->global(ierr) != gidJ ) { // recover
assert(0); // ?????
	  int cndf = sz/fndf; assert(sz%fndf==0); assert(par->global(ierr)<0);
	  par->SetNDF( cndf ); 
	  par->setGlobal( gidJ ); //par->setProc(p); par->setGEQ(g);
	}
	assert( par->global() == gidJ );
	if( (pDAP=par->DAP_) == NULL && sz > 0 ) { 
	  ierr = PetscMalloc(sz*sizeof(shpfloat),&pDAP); CHKERRQ(ierr); 
	  par->DAP_ = pDAP;
	  if( par->getPsz() == 0 ) par->SetPsz_hard( sz );
	  else assert( par->getPsz() == sz);
	}
	//else assert(par->DAP_ == NULL);
	shpfloat *fpb = (shpfloat*)pb; 
	for( ii = 0 ; ii < sz ; ii++ ) pDAP[ii] = fpb[ii]; // copy in
	pb = (int*)(fpb + sz);
      } 
    }
    assert(pb-buff == tt);    PetscFree(buff);
  }
  
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, gComm_, &ii, &status ); assert(!ii);

  return 0;
}

////////////////////////////////////////////////////////////////////////
// CG singular value helpers
////////////////////////////////////////////////////////////////////////
#undef __FUNCT__  
#define __FUNCT__ "PromLINPACKcgpthy"
#define	PROM_LP_MAX(a,b) (((a)>(b))?(a):(b))
#define	PROM_LP_MIN(a,b) (((a)<(b))?(a):(b))
/* PromLINPACKcgpthy **************************************
 *
 */   
static double PromLINPACKcgpthy(double *a,double *b)
{
  /* System generated locals */
  double ret_val,d__1,d__2,d__3;
  
  /* Local variables */
  double p,r,s,t,u;
  
  /*     FINDS DSQRT(A**2+B**2) WITHOUT OVERFLOW OR DESTRUCTIVE UNDERFLOW */

  /* Computing MAX */
  d__1 = fabs(*a),d__2 = fabs(*b);
  p = PROM_LP_MAX(d__1,d__2);
  if (!p) {
    goto L20;
  }
  /* Computing MIN */
  d__2 = fabs(*a),d__3 = fabs(*b);
  /* Computing 2nd power */
  d__1 = PROM_LP_MIN(d__2,d__3) / p;
  r = d__1 * d__1;
 L10:
  t = r + 4.;
  if (t == 4.) {
    goto L20;
  }
  s = r / t;
  u = s * 2. + 1.;
  p = u * p;
  /* Computing 2nd power */
  d__1 = s / u;
  r = d__1 * d__1 * r;
  goto L10;
 L20:
  ret_val = p;
  return ret_val;
} /* cgpthy_ */


/* tql1.f -- translated by f2c (version of 25 March 1992  12:58:56).
   By Barry Smith on March 27, 1994. 
   Eispack routine to determine eigenvalues of symmetric 
   tridiagonal matrix 

  Note that this routine always uses real numbers (not complex) even if the underlying 
  matrix is Hermitian. This is because the Lanczos process applied to Hermitian matrices
  always produces a real, symmetric tridiagonal matrix.
*/

#undef __FUNCT__  
#define __FUNCT__ "PromLINPACKcgtql1"
/* PromLINPACKcgtql1 *******************************
 * 
 */  

int PromLINPACKcgtql1( int *n, double *d, double *e, int *ierr )
{
  /* System generated locals */
  int    i__1,i__2;
  double d__1,d__2,c_b10 = 1.0;
  
  /* Local variables */
  double c,f,g,h;
  int    i,j,l,m;
  double p,r,s,c2,c3 = 0.0;
  int    l1,l2;
  double s2 = 0.0;
  int    ii;
  double dl1,el1;
  int    mml;
  double tst1,tst2;
  
  /*     THIS SUBROUTINE IS A TRANSLATION OF THE ALGOL PROCEDURE TQL1, */
  /*     NUM. MATH. 11, 293-306(1968) BY BOWDLER, MARTIN, REINSCH, AND */
  /*     WILKINSON. */
  /*     HANDBOOK FOR AUTO. COMP., VOL.II-LINEAR ALGEBRA, 227-240(1971). */
  
  /*     THIS SUBROUTINE FINDS THE EIGENVALUES OF A SYMMETRIC */
  /*     TRIDIAGONAL MATRIX BY THE QL METHOD. */
  
  /*     ON INPUT */
  
  /*        N IS THE ORDER OF THE MATRIX. */
  
  /*        D CONTAINS THE DIAGONAL ELEMENTS OF THE INPUT MATRIX. */
  
  /*        E CONTAINS THE SUBDIAGONAL ELEMENTS OF THE INPUT MATRIX */
  /*          IN ITS LAST N-1 POSITIONS.  E(1) IS ARBITRARY. */
  
  /*      ON OUTPUT */
  
  /*        D CONTAINS THE EIGENVALUES IN ASCENDING ORDER.  IF AN */
  /*          ERROR EXIT IS MADE, THE EIGENVALUES ARE CORRECT AND */
  /*          ORDERED FOR INDICES 1,2,...IERR-1, BUT MAY NOT BE */
  /*          THE SMALLEST EIGENVALUES. */
  
  /*        E HAS BEEN DESTROYED. */
  
  /*        IERR IS SET TO */
  /*          ZERO       FOR NORMAL RETURN, */
  /*          J          IF THE J-TH EIGENVALUE HAS NOT BEEN */
  /*                     DETERMINED AFTER 30 ITERATIONS. */
  
  /*     CALLS CGPTHY FOR  DSQRT(A*A + B*B) . */
  
  /*     QUESTIONS AND COMMENTS SHOULD BE DIRECTED TO BURTON S. GARBOW, */
  /*     MATHEMATICS AND COMPUTER SCIENCE DIV, ARGONNE NATIONAL LABORATORY 
   */
  
  /*     THIS VERSION DATED AUGUST 1983. */
  
  /*     ------------------------------------------------------------------ 
   */
  double ds;
  
  --e;
  --d;
  
  *ierr = 0;
  if (*n == 1) {
    goto L1001;
  }
  
  i__1 = *n;
  for (i = 2; i <= i__1; ++i) {
    e[i - 1] = e[i];
  }
  
  f = 0.;
  tst1 = 0.;
  e[*n] = 0.;
  
  i__1 = *n;
  for (l = 1; l <= i__1; ++l) {
    j = 0;
    h = (d__1 = d[l],fabs(d__1)) + (d__2 = e[l],fabs(d__2));
    if (tst1 < h) {
      tst1 = h;
    }
    /*     .......... LOOK FOR SMALL SUB-DIAGONAL ELEMENT .......... */
    i__2 = *n;
    for (m = l; m <= i__2; ++m) {
      tst2 = tst1 + (d__1 = e[m],fabs(d__1));
      if (tst2 == tst1) {
	goto L120;
      }
      /*     .......... E(N) IS ALWAYS ZERO,SO THERE IS NO EXIT */
      /*                THROUGH THE BOTTOM OF THE LOOP .......... */
    }
  L120:
    if (m == l) {
      goto L210;
    }
  L130:
    if (j == 30) {
      goto L1000;
    }
    ++j;
    /*     .......... FORM SHIFT .......... */
    l1 = l + 1;
    l2 = l1 + 1;
    g = d[l];
    p = (d[l1] - g) / (e[l] * 2.);
    r = PromLINPACKcgpthy(&p,&c_b10);
    ds = 1.0; if (p < 0.0) ds = -1.0;
    d[l] = e[l] / (p + ds*r);
    d[l1] = e[l] * (p + ds*r);
    dl1 = d[l1];
    h = g - d[l];
    if (l2 > *n) {
      goto L145;
    }
    
    i__2 = *n;
    for (i = l2; i <= i__2; ++i) {
      d[i] -= h;
    }
    
  L145:
    f += h;
    /*     .......... QL TRANSFORMATION .......... */
    p = d[m];
    c = 1.;
    c2 = c;
    el1 = e[l1];
    s = 0.;
    mml = m - l;
    /*     .......... FOR I=M-1 STEP -1 UNTIL L DO -- .......... */
    i__2 = mml;
    for (ii = 1; ii <= i__2; ++ii) {
      c3 = c2;
      c2 = c;
      s2 = s;
      i = m - ii;
      g = c * e[i];
      h = c * p;
      r = PromLINPACKcgpthy(&p,&e[i]);
      e[i + 1] = s * r;
      s = e[i] / r;
      c = p / r;
      p = c * d[i] - s * g;
      d[i + 1] = h + s * (c * g + s * d[i]);
    }
    
    p = -s * s2 * c3 * el1 * e[l] / dl1;
    e[l] = s * p;
    d[l] = c * p;
    tst2 = tst1 + (d__1 = e[l],fabs(d__1));
    if (tst2 > tst1) {
      goto L130;
    }
  L210:
    p = d[l] + f;
    /*     .......... ORDER EIGENVALUES .......... */
    if (l == 1) {
      goto L250;
    }
    /*     .......... FOR I=L STEP -1 UNTIL 2 DO -- .......... */
    i__2 = l;
    for (ii = 2; ii <= i__2; ++ii) {
      i = l + 2 - ii;
      if (p >= d[i - 1]) {
	goto L270;
      }
      d[i] = d[i - 1];
    } 
    
  L250:
    i = 1;
  L270:
    d[i] = p;
  }
   
  goto L1001;
  /*     .......... SET ERROR -- NO CONVERGENCE TO AN */
  /*                EIGENVALUE AFTER 30 ITERATIONS .......... */
 L1000:
  *ierr = l;
 L1001:
  return 0;
} /* cgtql1_ */

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix_base::ComputeLambda" 
/* PromMatrix_base::ComputeLambda ******************************************
 *
 *  INPUT:
 *    - A:
 *    - maxits:
 *    - lambda:
 *
 *  SIDE EFFECTS:
 *    - sets 'DRhoA_'
 *
 *  RETURN: 
 *     - PETSc error code 
 */
int PromMatrix_base::ComputeLambda(const PromOptions &opt,
				   PromPerfMonitor &perf, int maxits, 
				   PromPC_base *const pc/*=NULL*/) const
{
  int ii,ierr,tt,mype = map_.comm_.mype(); double *dd;
  const int myeq0=map_.my0Eq(),fact=isComplex()?2:1;
  const int nloceqf = fact*map_.nlocalEq();
  double emax, emin = (double)RAND_MAX/2.0, normMx,tol;
  const PromMatrix * const This = dynamic_cast<const PromMatrix*>(this);
  bool failed = isComplex(); 
  const PromMap *p_map = NULL; 
  PromVector_base *rr,*xx; PromCRVector *cr_r = NULL;

  PetscFunctionBegin; 
  // vectors
  if( isComplex() ) {
    rr = new PromComplexVector( map_ ); CHKNEW(rr,1);
    ierr = rr->Set(0.0); CHKERRQ(ierr); // for imag. part
    xx = new PromComplexVector( map_ ); CHKNEW(xx,1);
  }
  else {
    rr = new PromVector( map_ ); CHKNEW(rr,1);
    xx = new PromVector( map_ ); CHKNEW(xx,1);
  }

  // set random
  ierr = GetDiagonal( xx );      CHKERRQ(ierr);
  ierr = xx->NormMax( &normMx ); CHKERRQ(ierr);
  tol = normMx * 1.e-4;
  //if(tol<10.*PROM_RAP_SHIFT)tol=10.*PROM_RAP_SHIFT;
  //if(tol<normMx*PROM_DIAG_FACTOR)tol=normMx*PROM_DIAG_FACTOR;
  ierr = xx->GetArray( &dd ); CHKERRQ(ierr);
  for( ii = 0, tt = myeq0 ; ii<nloceqf ; ii += fact, tt++){
    double t = dd[ii], temp; 
    if( t < tol ) temp = 0.0;
    else temp = ((double)rand()-emin)/(2. * emin);
    ierr = rr->SetValues( 1, &tt, &temp, INSERT_VALUES ); CHKERRQ(ierr);
  }
  ierr = xx->RestoreArray( &dd );    CHKERRQ(ierr);
  //
  if( cr_r != NULL ) {
    ierr = cr_r->p_->Set( 0. );  CHKERRQ(ierr);
  }
  ierr = rr->Assembly(); CHKERRQ(ierr); // required for funking PETSc state issue

  // solver
  PromSolver *solver = new PromSolver(opt,perf,map_,p_map);  CHKNEW(solver,1);
  ierr = solver->SetTolerances(maxits,1.e-100,1.e-200,1.e5); CHKERRQ(ierr); 
  ierr = solver->SetKSPType( PROMKSPCG );                    CHKERRQ(ierr);
  ierr = solver->SetComputeSVD();                            CHKERRQ(ierr);
  if( pc == NULL ){
#if defined(PROM_USE_NODAL_SMAGG)
    ierr = solver->CreatePC( PROMPCNODALASM, xx );CHKERRQ(ierr);
#else 
    ierr = solver->CreatePC( PROMPCJACOBI, xx );CHKERRQ(ierr);
#endif 
    ierr = solver->SetOperator( This ); CHKERRQ(ierr); 
    ierr = solver->SetUp(); CHKERRQ(ierr);
  }
  else {
    ierr = solver->SetPC( pc, This ); CHKERRQ(ierr); 
  }
  // solve -- keep it quit
  solver->verbose_ = -1; // flag for compute rho(A) 
  // try CG
  if( !failed ) {
    ierr = solver->Solve( rr, xx, &ii, TRUE, TRUE );
    if( ii < 0 ) ii = -ii;
    if( ierr < 0 ) { CHKERRQ(ierr); } // real error
    else {
      solver->ComputeExtremeSVD( ii, emax, emin );
      if( ierr > 0 ){ SetIndefinite(); failed = TRUE; }
      else if( emin <= 0. ) SetIndefinite(); 
      else SetIndefinite( FALSE );
      if( emax > 10.0) failed = TRUE; // dont trust large vals -- try other
      else if( (ii==maxits || isIndefinite()) && opt.verbose_ > 0 ){
	PetscPrintf(map_.comm_(),"\t[%d]%s Warning its=%d/%d emax=%g -- %s\n",
		    mype,__FUNCT__,ii,maxits,emax,isIndefinite()?"*indfinite*":"");
      }
    }
  }
  else emax = 0.0;
  // gmres
  if( failed && FALSE ) { // don't trust gmres
    ierr = solver->SetTolerances( maxits );    CHKERRQ(ierr);
    ierr = solver->SetNEig( maxits );          CHKERRQ(ierr); 
    ierr = solver->SetKSPType( PROMKSPGMRES ); CHKERRQ(ierr);
    assert( solver->computeSVD() );
    //
    ierr = solver->Solve( rr, xx, &ii, TRUE, TRUE );    CHKERRQ(ierr);
    if( ii < 0 ) ii = -ii; 
    ierr = solver->ComputeExtremeSVD( ii, emax, emin );  CHKERRQ(ierr);
    CHKERRQ(ierr);
    if( emax < 1.0 || emax > 10. || TRUE ) failed = TRUE; 
    else failed = FALSE;
    if( emin <= 0. ) SetIndefinite(); 
    if( opt.verbose_> 0 ){
      PetscPrintf(map_.comm_(),"[%d]%s GMRES rho(A)=%g, emin=%g\n",
		  mype,__FUNCT__,emax,emin);
      
    }
  }
  // power
  if( failed ) { 
    if( opt.verbose_ > 1 ){
      PetscPrintf(map_.comm_(),"[%d]%s Power method (emax=%e)\n",
		  mype,__FUNCT__,emax);
    }
    ierr = solver->SetKSPType( PROMKSPPREONLY );       CHKERRQ(ierr);
    ierr = solver->SetTolerances( 1 );                 CHKERRQ(ierr);
    // 
    ierr = rr->Normalize( &emin );                     CHKERRQ(ierr);
    ierr = Mult( rr, xx );                             CHKERRQ(ierr);
    ierr = solver->Solve( xx, rr, &tt, TRUE, TRUE );   CHKERRQ(ierr);
    ierr = rr->Normalize( &emin );                     CHKERRQ(ierr);
    for( ii = 1 ; ii < maxits - 1 ; ii++ ) {
      ierr = Mult( rr, xx );                           CHKERRQ(ierr);
      ierr = solver->Solve( xx, rr, &tt, TRUE, TRUE ); CHKERRQ(ierr);
      emax = emin;
      ierr = rr->Normalize( &emin );                   CHKERRQ(ierr);
      if( opt.verbose_ > 1 && ii%5 == 0 ){
	PetscPrintf(map_.comm_(),"\t[%d] %d) power rho(A)=%g\n",mype,ii,emin);
      }
      if( emin < 1.01 * emax ) break;
    }
    emax = 1.05*emin; emin = 1.0;
    if( opt.verbose_ > 1 ){
      PetscPrintf(map_.comm_(),"[%d]%s power rho(A)=%g\n",mype,__FUNCT__,emax);
    }
    if( emax < 1.0 || emax > 15. ) emax = 15.0;
  }

  // clean up
  if( solver->pc_ == pc ) solver->pc_ = NULL; // hide from destructor (hack)
  delete solver;
  delete xx; delete rr;
  SetDRho( emax ); 
  if( opt.verbose_>1 || emin <= 0.0 || emax < 1.0 ){
    PetscPrintf(map_.comm_(),
		"[%d]%s Eigens=(%g,%g),%d/%d its, COND(A)=%g rho(A)=%g N=%d\n",
		mype,__FUNCT__,emin,emax,ii,maxits,emax/emin,DRhoA_,
		getM() + ((cr_r==NULL) ? 0 : cr_r->p_->getN()) );
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "PromSolver::SetComputeSVD"
/* PromSolver::SetComputeSVD **************************************
 *
 * CG amd gmres only 
 *
 */   
int PromSolver::SetComputeSVD()   
{
  assert(this!=NULL);
  int ierr;
  
  PetscFunctionBegin;
  if(ksp_type_!=PROMKSPCG&&ksp_type_!=PROMKSPGMRES&&ksp_type_!=PROMKSPFGMRES){
    SETERRQ1(1,"NO SingularValues ksp_type=%d",ksp_type_);
  }
  else if( maxits_ <= 0 ){
    PetscPrintf(PETSC_COMM_WORLD,"[%d] ERROR: maxit=%d\n",0,maxits_);
  }
  if( ksp_type_ == PROMKSPCG ) {
    if( origin_.e != NULL ){ ierr = PetscFree( origin_.e ); CHKERRQ(ierr); }
    if( maxits_ +1 > nEig_ ) nEig_ = maxits_ + 1;
    ierr = PetscMalloc( 4*nEig_*sizeof(double), &origin_.e ); CHKERRQ(ierr);
    d_.e = origin_.e + nEig_; ee_.e = origin_.e + 2*nEig_; 
    dd_.e = origin_.e + 3*nEig_;
  }
  // gmres do all the time
  PetscFunctionReturn(0);
}
#undef __FUNCT__  
#define __FUNCT__ "PromSolver::ComputeExtremeSVD"
/* PromSolver::ComputeExtremeSVD *******************************
 * 
 */  
int PromSolver::ComputeExtremeSVD(int nits, double &emax, double &emin)
{
  int ierr;

  if( nits <= 1 ) emax = emin = 1.0; // none (or 1) to do
  else if( ksp_type_ == PROMKSPCG ) {
    ierr = ComputeExtremeSVD_cg( nits, emax, emin ); CHKERRQ(ierr); 
  }
  else if( ksp_type_ == PROMKSPGMRES || ksp_type_ == PROMKSPFGMRES ) { 
    ierr = ComputeExtremeSVD_gmres( nits, emax, emin ); CHKERRQ(ierr); 
  }
  else emax = emin = 1.0; // error
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromSolver::ComputeExtremeSVD_cg"
/* PromSolver::ComputeExtremeSVD_cg *******************************
 * 
 */  
int PromSolver::ComputeExtremeSVD_cg(int nn, double &emax, double &emin)
{
  int j;
  double *d, *e, *dd, *ee;  assert(nn<=maxits_);
  d = d_.e;  e = origin_.e;  dd = dd_.e;  ee = ee_.e;
  /// copy tridiagonal matrix to work space 
  for ( j = 0; j < nn ; j++) { ee[j] = e[j]; dd[j] = d[j]; }
  PromLINPACKcgtql1( &nn, dd, ee, &j );
  if (j != 0) SETERRQ(1,"Error from tql1(); eispack eigenvalue routine");  
  emin = dd[0]; emax = dd[nn-1];

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromSolver::ComputeExtremeSVD_gmres"
/* PromSolver::ComputeExtremeSVD_gmres *******************************
 * 
 */  
int PromSolver::ComputeExtremeSVD_gmres( int its, double &emax, double &emin )
{
#define PROM_SVD 4096
#define PROM_SVD2 8192
  int max_k=nEig_,LDA=max_k+2,ii,it;
  double R__[PROM_SVD2],*R =R__, tmp; R__[PROM_SVD2-1] = -98.;

  it = its%max_k;
  if( it == 0 ) it = max_k;
  const int n = it; const bool isCmplx = A_->isComplex();
  if( n > PROM_SVD || n <= 1){ emax = 1.; emin = 1.; }
  else {
    // just get the diagnal for eigenvalues
    for( emax=-1.e99,emin=1.e99,ii=0 ; ii<n ; ii++ ){
      it = ii*LDA + ii; 
      if( !isCmplx ) tmp = origin_.e[it];
#if defined(PROM_HAVE_COMPLEX)
      else tmp = origin_.c[it].real();
#endif
      if( tmp > emax ) emax = tmp;
      else if( tmp < emin ) emin = tmp;
      R[ii] = tmp;
    }
  }
  /*else {
    int bsz = n*LDA*sizeof(double);
    if( LDA*LDA > PROM_SVD2 ) {
    int ierr = PetscMalloc( bsz, &R );  CHKERRQ(ierr); 
    }
    int lwork=5*PROM_SVD,work[5*PROM_SVD],idummy=LDA; double sdummy,realpart[PROM_SVD];
    // copy R matrix to work space
    if( A_->isComplex() ) {
    #if defined(PROM_HAVE_COMPLEX)
    for(ii=0;ii<n*LDA;ii++) R[ii] = origin_.c[ii].real();
    #else
    SETERRQ(1,"complex matrix w/o PROM_HAVE_COMPLEX");
    #endif
    }
    else{
    ierr = PetscMemcpy( R, origin_.e, bsz );CHKERRQ(ierr);
    }
    // zero below diagonal garbage 
    for (ii=0; ii<n; ii++) R[ii*LDA+ii+1] = 0.0;
    #if defined(PARCH_rs6000)
    int iopt = 10;
    dgesvf_( &iopt, R, &LDA, &sdummy, &iopt, &iopt, 
    realpart, &n, &n, work,& lwork );
    #else
    dgesvd_("N","N",&n,&n,R,&LDA,realpart,&sdummy,&idummy,&sdummy,&idummy,work,&lwork,&ierr);
    CHKERRQ(ierr);
    #endif
    emax = realpart[0];  emin = realpart[n-1]; 
    assert( n < 2 || realpart[n-1] <= realpart[n-2]);
    //
    if( R != R__ ) PetscFree(R);
    }*/
  assert(R__[PROM_SVD2-1] == -98.);
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::SmoothScalarP" 
/*  PromGrid::SmoothScalarP ***********************************
 *
 *   - use native PETSc methods for smoothing
 *
 *   INPUT:
 *    - nextG: other version of this grid
 *
 *   SIDE EFFECTS:
 *    - changes the prolongators matrix for 'nextG'
 * 
 *   RETURN:
 *     - PETSc error code
 */
int PromGrid::SmoothScalarP( PromGrid *nextG ) const 
{
  int ierr; 
  Mat P = nextG->prolongator_->mat_, A0 = stiffness_->mat_;
  
  perf_mon_.EventBegin( PromContext::prom_logs_[LOG_MAPS] );
  const PromMatrix *AA = stiffness_; assert(AA!=NULL);
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS0]);  
#if !defined(PROM_NO_STAGES)
  perf_mon_.StagePush(PromContext::prom_stages_[5]);
#endif
  // get lambda - collective on 'wcomm'
  ierr = AA->ComputeLambda( options_, perf_mon_, 20 ); CHKERRQ(ierr);
  double lambda = AA->getDRho();
#if !defined(PROM_NO_STAGES)
  perf_mon_.StagePop();
#endif
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS0]);
  // form D: -1.5/lambda D^-1
  PromVector diag( AA->map_, AA->isComplex() ? 2 : 1 );
  ierr = AA->PromMatrix::GetDiagonal( &diag );  CHKERRQ(ierr);
  ierr = diag.Reciprocal(); CHKERRQ(ierr);
  ierr = diag.Scale( -1.5 / (lambda*1.1) ); CHKERRQ(ierr);
  // form AP
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_MIS1]);
  Mat dap; 
  ierr = MatMatMult( A0, P, MAT_INITIAL_MATRIX, 3.0, &dap ); CHKERRQ(ierr);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_MIS1]);
  // form DAP
  PromMatrix mDAP( AA->map_, dap ); mDAP.malled_ = false;
  ierr = mDAP.Scale( &diag, NULL ); CHKERRQ(ierr);
  // form P1 = P_0 - DAP_0 (dap += P)
  PetscInt end, my0, ncols, ii; const PetscInt *cols; const PetscScalar *vv;
  ierr = MatGetOwnershipRange( P, &my0, &end ); CHKERRQ(ierr);
  for( ii = my0 ; ii < end ; ii++ ) {
    ierr = MatGetRow( P, ii, &ncols, &cols, &vv ); CHKERRQ(ierr);
    ierr = MatSetValues( dap, 1, &ii, ncols, cols, vv, ADD_VALUES );
    ierr = MatRestoreRow( P, ii, &ncols, &cols, &vv ); CHKERRQ(ierr);
  }
  ierr = MatAssemblyBegin( dap, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
  ierr = MatAssemblyEnd( dap, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
  // hack new prolongator in!
  nextG->prolongator_->mat_ = dap; ierr = MatDestroy(P); CHKERRQ(ierr);
  perf_mon_.Flops(1); // make sure this is not in symbolic
  perf_mon_.EventEnd( PromContext::prom_logs_[LOG_MAPS] );
 
  return 0;
}
