//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: geo_coar_move.C,v 1.2 2004/08/20 07:19:03 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           geo_coar_move.C
//
#include "prom_grid.hh"

#define MAXFACETS 50
static PromList<PromStack*> *list = NULL;  // 
static PromList<PromStack*> *nodelist = NULL;         

/* pivot34() ******************************************
 */
static inline double **pivot34( const int j , double a[3][4] ) 
{
  double temp, max = fabs(a[j][j]);
  int jj, i;  
  for ( jj = j, i = j + 1 ; i < 3 ; i++ ){
    temp = fabs(a[i][j]);
    if( temp > max ) {
      jj = i;
      max = temp;
    }
  }
  if ( jj == j ) return NULL;
  else {
    // swap rows j & jj      swap4 ( a[jj], a[j]);  
    double *aa = a[jj], *bb = a[j];
    double temp0 = aa[0];  aa[0] = bb[0];  bb[0] = temp0;
    double temp1 = aa[1];  aa[1] = bb[1];  bb[1] = temp1;
    double temp2 = aa[2];  aa[2] = bb[2];  bb[2] = temp2;
    double temp3 = aa[3];  aa[3] = bb[3];  bb[3] = temp3;
    return (double **)a;
  }
}  

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::MoveNodes"
/* PromGeoGrid::MoveNodes() ******************************************
 *
 *  - move nodes in 'list' of adjac nodes
 *
 *  INPUT:
 *    - numUnfound: number of nodes to cover (in/out)
 *    - maxloops: max number of iterations  
 *    - maxoff: max off shape value. (out)
 *
 *  SIDE EFFECTS:
 *    - adds orphaned nodes to interior elements
 *    - sets adjacentcies in node's 'list' - just like the end of constructor1
 *    - 'marks' nodes that are on exterior now
 *    - uses tags 20,21 
 *    - collective
 *
 *  RETURN:
 *    - PETSc error code.
 */
int PromGeoGrid::MoveNodes( int &numUnfound, const int maxloops, 
			    gfloat &maxoff, PromList<PromPoly*> *elemList )
{
  PromGeoNode *curr, *edges[MAXFACETS][2];
  PromList<PromStack*> unfound, edgeLists[MAXFACETS], edgeUnfound;
  int                 j, nEdgeNd, ierr, xx;
  const int           todo = numUnfound;

  if( !list ) { list = new PromList<PromStack*>; CHKNEW_void(list,1); }
  if(!nodelist){nodelist = new PromList<PromStack*>;CHKNEW_void(nodelist,1);}

  for( xx = j = 0 ; xx < nNodes_ ; xx++ ){
    curr = (PromGeoNode*)nodes_[xx]; 
    assert(!curr->isMarked());
    j += curr->numChildren();
  }
  assert(j==numUnfound);
  
  // iterate til done
  int iter1 = 0, last2 = 0, last1 = numUnfound, lastrun = 0;
  while( iter1++ < maxloops ) {    
    int keepgoing = !(last2 == numUnfound || numUnfound == 0); 
    MPI_Allreduce( &keepgoing, &j, 1, MPI_INT, MPI_SUM, gComm_ ); 
    if( j == 0 || iter1 == maxloops ) lastrun++;
    else lastrun = 0;
    if( lastrun == 2 ) break; // two loop2
    
    last2 = last1;  
    last1 = numUnfound;
    const bool stalled = ( last2 - last1 < 1 && iter1 > 1 );
    if(options_.verbose_>1)PetscPrintf(gComm_,"\n[%d]Move(%d/%d) %d/%d nodes -- %s",
		comm_.mype(), iter1, maxloops, todo - numUnfound, todo,
		lastrun ? "lastrun" : stalled ? "stalled" : "");
    // exterior nodes only - DO NOT MOVE GHOSTS
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
      //for( xx = nLocalNd_-1 ; xx >= 0 ; xx-- ){
      curr = (PromGeoNode*)nodes_[xx];  
      if(!curr->isExterior()||curr->symbData_->elemList.isEmpty())curr->mark();
      if( curr->isMarked() ){ assert( curr->isChildListEmpty() ); continue; }
      if( iter1 == 1 && curr->getTopoType() != PROM_SURFACE ) continue;
      // skip boundary moves on last
      if( lastrun && curr->isBound() ) continue; 
      gfloat scale = 0.0; PromCoord deltaVec(0.0, 0.0, 0.0); 
      bool noop = FALSE;
      // get list of facets
      ierr = GetListOfFacets( curr, edges, nEdgeNd, edgeLists, unfound,
			      edgeUnfound, deltaVec, scale, noop );
      CHKERRQ(ierr);
      if ( noop ) continue;
      noop = FALSE;
      if ( scale > 0.0 ){
	ierr = MoveNode( curr, deltaVec, scale, elemList );    CHKERRQ(ierr);
	if(options_.verbose_>1)PetscPrintf(gComm_,".");
	if( scale == 0.0 ) noop = TRUE;
      }
  
      // reset unfound nodes
      ierr = ResetUnfound_nodes( curr, edges, nEdgeNd, edgeLists, unfound, 
				  edgeUnfound, numUnfound, (scale == 0.0), 
				  stalled ); 
      CHKERRQ(ierr);
      assert( curr->numChildren() <= numUnfound );
      if( scale == -1.0 ) {
	//curr->mark(); 	ierr = EmptyChildren( curr );    CHKERRQ(ierr);
	//if(options_.verbose_>1)PetscPrintf(gComm_,"X"); 
      }
      else if( !curr->isChildListEmpty() ) {      
	//ierr = EmptyChildren( curr );          CHKERRQ(ierr);
      }
    } // nodes
    assert( numUnfound >= 0 );
    
    // receive updates on coodinates
    MPI_Barrier(gComm_);  // not garanteed to work!!!
    ierr = CheckForUpdate( TRUE, elemList );                   CHKERRQ(ierr);
  } 
  MPI_Barrier(gComm_);  // not garanteed to work!!!
  ierr = CheckForUpdate( FALSE, elemList  );                    CHKERRQ(ierr);

  // clean up
  list->removeAll(); nodelist->removeAll();
  delete list; list = NULL; 
  delete nodelist; nodelist = NULL; 
  for( xx = 0 ; xx < nNodes_ ; xx++ ){
    curr = (PromGeoNode*)nodes_[xx]; 
    curr->mark(FALSE);
  }  

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::CheckForUpdate"
/* PromGeoGrid::CheckForUpdate() ******************************************
 *
 *  - receive ghost node coordinate updates
 *
 *  INPUT:
 *   respond: 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::CheckForUpdate( bool respond, PromList<PromPoly*> *elemList ) 
{
  int          receive,id,ierr,ii; bool flag;
  MPI_Status   status;
  PromCoord    offset;
  PromGeoNode   *curr;
  double lsbuf[PROM_BSZ], lrbuf[PROM_BSZ]; 

  while(1) {
    MPI_Iprobe( MPI_ANY_SOURCE,20,gComm_, &receive, &status);
    if( !receive ){
      MPI_Iprobe( MPI_ANY_SOURCE,21,gComm_, &receive, &status);
    }
    if( !receive ) break;
    
    // receive < coord, localid > and get node
    MPI_Get_count( &status, MPI_INT, &ii ); 
    if( ii > PROM_BSZ ) SETERRQ(1,"ERROR");
    MPI_Recv( lrbuf, PROM_BSZ, MPI_DOUBLE, status.MPI_SOURCE, status.MPI_TAG,
	      gComm_, &status );
    int *pb = (int*)lrbuf; 
    offset.Read( &pb ); PromContext::Read( id, pb );    assert( ii == pb-(int*)lrbuf );
    curr = (PromGeoNode*)nodes_[id];    assert( id < nNodes_ );
    assert( curr->isGhost() || status.MPI_TAG == 21 );
    // move it                    ->symbGeoData_
    curr->symbGeoData_->coord += offset;
    assert( list->isEmpty() );
    int isundone = FALSE;
    assert( !curr->symbData_->elemList.isEmpty() );
    ierr = ResetChildElems( &curr->symbData_->elemList, list, elemList, 
			    &flag, FALSE );   
    if(ierr)PetscPrintf(PETSC_COMM_SELF,"TAG=%d bad=%d\n",status.MPI_TAG,flag);
    CHKERRQ(ierr); 
    if( !list->isEmpty() || flag ) { // undo move and  reset 
      // this is bad - an elem is probably insideout
      curr->symbGeoData_->coord -= offset;
      ierr = ResetChildElems( &curr->symbData_->elemList, list, elemList,
			      &flag, FALSE );   CHKERRQ(ierr);
      if( !list->isEmpty() || flag ) {
	PetscPrintf(PETSC_COMM_SELF,"[%d]ERROR:%d\n",comm_.mype(),curr->index_());
	SETERRQ(1,"Unable to recover from in-side-out elem\n");
      }
      curr->mark(); ierr = EmptyChildren( curr );                CHKERRQ(ierr);
      PetscPrintf(gComm_,"E");
      isundone = TRUE;
    } 
    if( status.MPI_TAG == 20 ) {  
      assert( curr->isGhost() );
      if( respond && isundone ) {
	offset = offset * -1.0;
	// send undo                  
	pb = (int*)lsbuf; offset.Write( &pb ); 
	PromContext::Write( curr->symbData_->ghostID.local(), pb ); 
	assert( curr->symbData_->ghostID.proc()/factor_ == status.MPI_SOURCE );
	// send err message < coord, localid >
	MPI_Send(lsbuf,pb-(int*)lsbuf,MPI_INT,status.MPI_SOURCE,21,gComm_);
      }
    }
    else if( isundone ) {
      if(options_.verbose_>1)PetscPrintf( PETSC_COMM_SELF,
				"\n\n[%d]WARNING bound %d NOT consistant\n",
				comm_.mype(), curr->index_() );
    }
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::GetListOfFacets"
/* PromGeoGrid::GetListOfFacets() ******************************************
 *
 *  - get facets of this node
 *
 *  INPUT:
 *    - curr: current node to move 
 *    - edges: array of node pairs - edges. (out)
 *    - nEdgeNd: size of edges (out)
 *    - edgeLists: array of list to put node children in (out)
 *    - unfound: list of nodes,from 'curr', to check for. (out)
 *    - edgeUnfound: list of nodes, from edge, to check for. (out)
 *    - deltaVec: vector to move node on, length is twice area of patch (out)
 *    - scale: scale of 'deltaVec' to move node. (out)
 *    - noop: flag for nothing to do. (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::GetListOfFacets( PromGeoNode *curr, PromGeoNode *edges[][2], 
				  int &nEdgeNd,
				  PromList<PromStack*>edgeLists[],
				  PromList<PromStack*> &unfound,
				  PromList<PromStack*> &edgeUnfound,
				  PromCoord &deltaVec, gfloat &scale, 
				  bool &noop )
{
  int ierr, nsurfs, surf[MAXFACETS][5], topfacets[5], num;;
  assert( edgeUnfound.isEmpty() && unfound.isEmpty() ); 
  assert( !curr->isGhost() );
  
  gfloat totvol = 0.0, tt;
  PromFacet  *facet, *facets[MAXFACETS];
  // get facets of this node (and edges of this patch)
  ierr = curr->MakeExteriorFacets( facets, edges, &totvol, this, nEdgeNd );
  CHKERRQ(ierr);
  // a covered exterior node
  if( nEdgeNd < 3 || totvol == 0.0 ) {
    assert( nEdgeNd == 0 || nEdgeNd == 50 ); 
    for ( int ii = 0 ; ii < nEdgeNd ; ii++ ) {
      Kill(facets[ii]); facets[ii] = NULL; 
    }
    curr->mark();  ierr = EmptyChildren( curr );                CHKERRQ(ierr);
    noop = TRUE;  
    if(options_.verbose_>1) PetscPrintf(gComm_,"Q");
    return 0;
  }
  else noop = FALSE;
  
  // check for multiple face
  ierr = GetFaces( curr, edges, nEdgeNd, surf, topfacets, nsurfs ); 
  CHKERRQ(ierr);
  if( nsurfs == 1 ) {    
    // get list, and remove done nodes
    curr->removeChildren ( &unfound );
    // get near by children
    num = getNeighborChildren( curr, edgeLists, edges, nEdgeNd );
    if ( unfound.isEmpty() && !num ) { // nothing to do here 
      for ( int ii = 0 ; ii < nEdgeNd ; ii++ ) {
	Kill(facets[ii]); facets[ii] = NULL; 
      }
      assert( edgeUnfound.isEmpty() && unfound.isEmpty() ); 
      noop = TRUE; return 0;
    }
  }
  else {
    for ( int ii = 0 ; ii < nEdgeNd ; ii++ ) {
      Kill(facets[ii]); facets[ii] = NULL; 
    }
    curr->mark(); ierr = EmptyChildren( curr );                CHKERRQ(ierr);
    if(options_.verbose_>1) PetscPrintf(PETSC_COMM_SELF,"S+(%d)",comm_.mype()); 
    scale = -1.0; return 0; //flag
  } 
  
  // make move vector and form edge's list, consolodate edgelists
  int ii, jj; 
  for ( ii = jj = 0 ; ii < nEdgeNd ; ii++ ) {
    facet = facets[ii]; 
    if( !facet->nodes_[0]->isGhost() && !facet->nodes_[1]->isGhost() && 
	!facet->nodes_[2]->isGhost() ){
      deltaVec += facet->norm_; jj++;
    }
    edgeUnfound.addHead( &edgeLists[ii] ); // add to big extra list
  }
  assert ( edgeUnfound.getCount() == num );
  if( jj == 0 ) { 
    for ( int ii = 0 ; ii < nEdgeNd ; ii++ ) {
      Kill( facets[ii] ); facets[ii] = NULL; 
    }
    scale = -1.0; return 0; // flag for a bad one
  }
  
  // check for valid vect 
  scale = 0.0, tt = deltaVec.length();
  for( ii = 0 ; ii < nEdgeNd ; ii++ ) {
    facet = facets[ii]; 
    while( facet->norm_ * deltaVec < 1.e-3 * tt * facet->norm_.length() ) {
      for ( int ii = 0 ; ii < nEdgeNd ; ii++ ) {
	Kill( facets[ii] ); facets[ii] = NULL; 
      }
      scale = -1.0; return 0; // flag for a bad one
    }
  }
  // get scale for vect
  ierr = GetScaleFactor( curr, deltaVec, &unfound, &edgeUnfound, edges, 
			   nEdgeNd, scale );
  CHKERRQ(ierr);
  
  // done with facets
  for( ii = 0 ; ii < nEdgeNd ; ii++ ) {
    Kill( facets[ii] ); facets[ii] = NULL; 
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::MoveNode"
/* PromGeoGrid::MoveNode() ******************************************
 *
 *  - move a node
 *
 *  INPUT:
 *    - curr: current node to move 
 *    - deltaVec: vector to move node on
 *    - scale: scale of 'deltaVec' to move node (in/out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::MoveNode( PromGeoNode *curr, const PromCoord &deltaVec, 
			   gfloat &scale, PromList<PromPoly*> *elemList )
{
  int ierr, kk ; bool bad; assert(scale > 0.0);
  assert( !curr->isGhost() );

  PromCoord offset( deltaVec * scale ); 
  // only move node in non-fixed direction
  curr->symbGeoData_->coord += offset;
  // reset fine nodes as these elemts were changed
  assert( list->isEmpty() );
  ierr = ResetChildElems( &curr->symbData_->elemList, list, elemList,
			  &bad, FALSE ); CHKERRQ(ierr);
  for( kk = 0 ; kk < 4 && bad ; kk++ ){
    curr->symbGeoData_->coord -= offset;
    offset *= 0.5; scale *= 0.5;      
    curr->symbGeoData_->coord += offset;
    ierr = ResetChildElems(&curr->symbData_->elemList, list, elemList,
			   &bad, FALSE);   CHKERRQ(ierr);
  }
  if ( !list->isEmpty() || bad ) {
    // undo move and  reset 
    curr->symbGeoData_->coord -= offset;
    scale = 0.0;                                           
    ierr = ResetChildElems( &curr->symbData_->elemList, list, elemList,
			    &bad, FALSE );  CHKERRQ(ierr);
    if ( !list->isEmpty() || bad ) {
      PetscPrintf(PETSC_COMM_SELF,"[%d]%s %d\n",comm_.mype(),__FILE__,curr->index_());
      SETERRQ(1,"Unable to recover from in-side-out elem - MoveNodes\n");
    }
  }
  else if ( curr->isBound() ) {
    // send updates of a move
    gfloat lsbuf[12]; int *pb;
    Position pos = curr->symbData_->dependents->getHeadPosition(); 
    assert(pos!=NULL);
    while ( pos ) {
      PromGlobalID *dep = curr->symbData_->dependents->getNext( pos ); 
      pb = (int*)lsbuf; offset.Write( &pb ); 
      PromContext::Write( dep->local(ierr), pb );
      CHKERRQ(ierr);
      // send(tag < coord, localid >
      MPI_Send(lsbuf, pb-(int*)lsbuf, MPI_INT, dep->proc()/factor_, 20,gComm_);
    }
  }
  assert( list->isEmpty() ); 

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::ResetUnfound_nodes"
/* PromGeoGrid::ResetUnfound_nodes() ******************************************
 *
 *  - clean up after move
 *
 *  INPUT:
 *    - curr: current node to move 
 *    - edges: array of node pairs - edges. 
 *    - nEdgeNd: size of edges 
 *    - edgeLists: array of list to put node children in 
 *    - unfound: list of nodes,from 'curr', to check for. 
 *    - edgeUnfound: list of nodes, from edge, to check for.
 *    - numUnfound: number of unfound nodes total. (in/out)
 *    - zeroScale: was the scale zero;
 *    - stalled: flag that we are slowing down, so try harder.
 *
 *  SIDE EFFECTS:
 *    - puts children back int nodes. 
 *
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::ResetUnfound_nodes(PromGeoNode *curr, PromGeoNode *edges[][2],
				    const int nEdgeNd,
				    PromList<PromStack*> edgeLists[], 
				    PromList<PromStack*> &unfound, 
				    PromList<PromStack*> &edgeUnfound,
				    int &numUnfound, const bool zeroScale, 
				    const bool stalled )
{
  PromList<PromStack*> *plist;
  int         kk, ii,mype=comm_.mype();
  Position    pos;
  PromGeoNode  *n1;

  // put nodes back:
  for( kk = 0, plist = &unfound ; kk < 2 ; kk++, plist = &edgeUnfound ){
    while ( !plist->isEmpty() ) {    
      PromGeoNode *targ = (PromGeoNode*)plist->removeHead();    
      // get original parent for edge nodes
      int edgei = -9;
      if ( kk == 1 ) {
	for ( ii = 0 ; ii < nEdgeNd ; ii++){
	  if ( (pos = edgeLists[ii].find(targ)) ) {
	    edgeLists[ii].removeAt( pos );
	    edgei = ii;
	    break;
	  }
	}
	n1 = edges[edgei][0];     assert( n1!=NULL && edgei != -9 );           
	//assert( !n1->isGhost() );
      }
      // try to put in an element
      bool flag = FALSE; // look in elements
      if ( !curr->symbData_->elemList.isEmpty() 
	   && targ->setElemWithList( &curr->symbData_->elemList, parBank_, 
				     mype,1, 1.e-4, TRUE,FALSE) ) {
	flag = TRUE; 
      }
      // resolve node
      if ( flag ) {                    // found  
	numUnfound--;       	assert( targ->symbGeoData_->parElem!=NULL );
      }
      else {                            // not found
	// get node (n1) to add 'targ' back into
	if ( kk == 0 ) n1 = curr; 
	// add back to a node
	assert( n1 && !n1->symbData_->elemList.isEmpty() );
	assert( !n1->isGhost() && !n1->isMarked() );
	n1->addChild( targ );
      }
    }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::ResetChildElems"
/* PromGeoGrid::ResetChildElems() ******************************************
 *
 *  - take union of input elem's (fine) nodes and find which elems they are
 *    in.  called after a node or element is altered (not nested).
 *
 *  INPUT:
 *    - elList: list of continuous elements to reset child nodes  
 *    - unfound: list of extra nodes, and list to return lost nodes. (in/oot)
 *    - badout: was an element bad, killed if "kill" (out)
 *    - kill: kill elements if they are bad
 *
 *  SIDE EFFECTS:
 *    - resetsets fines nodes list of these elems ("elemList").
 *    - resets nodes shape values.  
 *
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::ResetChildElems( PromList<PromPoly*> *elList, 
				  PromList<PromStack*> *unfound,
				  PromList<PromPoly*> *elemList,
				  bool *badout/*=NULL*/, 
				  const bool kill/*=FALSE*/)
{ 
  PromDTet     *elem;
  PromGeoNode  *node;
  Position      pos, pos2; assert( !elList->isEmpty() && nodelist->isEmpty() );
  bool bad = FALSE; int num,mype=comm_.mype(); 

  // collect all the nodes of these elements  
  pos = elList->getHeadPosition(); assert(pos!=NULL);
  while( (pos2=pos) ) {
    elem = (PromDTet*)elList->getNext( pos );
    elem->removeChildren ( nodelist );    
    // bail out if an inverted elem
    elem->calcTestData();
    if( elem->getVolume() <= 0.0 ) {
      bad = TRUE;
      if( kill ) {
	elList->removeAt( pos2 ); // this will cause an error message in ret.
	if( elem->willStrandNode(num) || num ) {
	  PetscPrintf(PETSC_COMM_SELF,"[%d]WARNING: NOT killing elem\n",comm_.mype());
	  SETERRQ(1,"NOT killing elem");
	}
	else {
	  retireElement( elem );
	  if(options_.verbose_>1)PetscPrintf(PETSC_COMM_SELF,
				    "[%d]***WARNING:killing elemen\n",comm_.mype());
	}
      }
      else if(options_.verbose_>1)  PetscPrintf( PETSC_COMM_SELF, badout ? "+" : "-");
    }
  }
  if( badout ) *badout = bad;

  // check for a nightmare
  if( kill && elList->isEmpty() ) {
    const int num = unfound->getCount() + nodelist->getCount();
    unfound->addTail( nodelist );
    assert( unfound->getCount() == num ); 
    nodelist->removeAll();
    return 0;
  }
  else if( bad ){
    const int num = unfound->getCount() + nodelist->getCount();
    unfound->addTail( nodelist );
    assert( unfound->getCount() == num ); 
    nodelist->removeAll();
    return 0;
  }
  
  // add extra nodes
  while ( !unfound->isEmpty() ) nodelist->addTail( unfound->removeHead() );
  
  // for all nodes - search elems for nodes
  while( !nodelist->isEmpty() ) {
    node = (PromGeoNode*)nodelist->removeHead();
    if( elList->isEmpty() ) SETERRQ(1,"lost all elements");
    node->symbGeoData_->parElem = NULL; // setElemWithList!!!
    elem = node->setElemWithList( elList, parBank_, mype, 1, 1.e-4, TRUE, FALSE );
    if( !elem ) {
      if(options_.verbose_>1)PetscPrintf(PETSC_COMM_SELF,"[%d]%s hard\n",comm_.mype(),__FUNCT__);
      pos = elemList->getHeadPosition();
      while( pos && !elem ) {
	elem = (PromDTet*)elemList->getNext( pos );  
	elem = node->setElem( elem, parBank_, mype, 1.e-4, TRUE, 1, NULL, FALSE );
      }
      if( !elem ){
	unfound->addTail( node );
	//assert(node->symbGeoData_->parElem==NULL);  
	if(options_.verbose_>1)
	  PetscPrintf(PETSC_COMM_SELF,"[%d]%s !add\n",comm_.mype(),__FUNCT__);
      }
    }
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::getNeighborChildren"
/* PromGeoGrid::getNeighborChildren() ****************************************
 *
 *  - get lists of facet children
 *
 *  INPUT:
 *    - curr: node at center of facets 
 *    - edgeLists: array of list to put node children in
 *    - edges: arrays with edge nodes
 *    - numfacets: number of facets (could be zero)
 *
 *  SIDE EFFECTS:
 *    - REMOVES some children for nodes child'ss list
 *
 *  RETURN:
 *    - number of children
 */
int PromGeoGrid::getNeighborChildren( const PromGeoNode *curr, 
				      PromList<PromStack*> edgeLists[], 
				      PromGeoNode *edges[][2],
				      const int numfacets )
{
  int count = 0;  
  gfloat targlen, len;    assert( !curr->isGhost() ); 
    
  for ( int ii = 0 ; ii < numfacets ; ii++ ) {
    PromGeoNode *n1 = edges[ii][0];
    PromList<PromStack*> *vlist = &edgeLists[ii];
    assert( vlist->isEmpty() );
    n1->removeChildren( vlist );
    PromCoord vec = n1->symbGeoData_->coord - curr->symbGeoData_->coord;
    len = vec.length();  
    // put far away nodes back 
    Position pos2, pos = vlist->getHeadPosition();
    while ( (pos2=pos) ) { 
      PromGeoNode *targ = (PromGeoNode*)vlist->getNext( pos );  
      vec = targ->symbGeoData_->coord - curr->symbGeoData_->coord;
      targlen = vec.length();
      //assert(!n1->isGhost());
      if ( targlen > len ) {
	vlist->removeAt( pos2 ); // undo if to far away
	n1->addChild( targ );
      }
    }
    count += vlist->getCount();
  }

  return count;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::GetFaces"
/* PromGeoGrid::GetFaces() ******************************************
 *
 *  - get exterior facets from list of facets
 *
 *  INPUT:
 *    - curr: node that is at center of facets.
 *    - edges: arrays with edge nodes.
 *    - numfacets: number of facets.
 *    - surf: array if facets in faces (out)
 *    - topfacets: number of facets in each face (out)
 *    - nsurfs: number of surfaces - should be one. (out)
 *
 *  SIDE EFFECTS:
 *    - 
 *
 *  RETURN:
 *    - PETSc error code.
 */
int PromGeoGrid::GetFaces( PromGeoNode *, PromGeoNode *edges[][2], 
			   const int numfacets, int surf[MAXFACETS][5], 
			   int topfacets[5], int & nsurfs)
{  
  PromGeoNode   *n1, *n2, *n3;
  int          done[MAXFACETS], numhit = 0, ii, starti, bad = FALSE;
  PetscMemzero( (char*)done, MAXFACETS * sizeof(int) );
  assert( numfacets > 2 );

  for( nsurfs = 0 ; nsurfs < 2 && numhit < numfacets ; nsurfs++ ) {  
    // get a continuous face
    for ( ii = 0, starti = -9 ; ii < numfacets && starti == -9 ; ii++ ) 
      if ( !done[ii] ) starti = ii;
    assert(starti != -9 );
    
    // get a face by fallowing pointers around
    int index = 0;
    n3 = n2 = edges[starti][1];
    do {
      for ( ii = 0 ; ii < numfacets ; ii++ ) {
        if ( done[ii] ) continue;
        n1 = edges[ii][0];
        if( n2 == n1 ) {
          surf[index][nsurfs] = ii;
          done[ii] = 1; // this is done
          break;
        }
      }
      if(n2 != n1){
	bad++; 
	if(options_.verbose_>1) PetscPrintf(PETSC_COMM_SELF,"[%d]%s() WARNING Bad\n",
				  comm_.mype(),__FUNCT__);
      }
      else {
	n2 = edges[ surf[index][nsurfs] ][1];
	index++;
	numhit++;
      } 
    }while( n2 != n3 && !bad );
    
    topfacets[nsurfs] = index;
  }
  assert( !bad || nsurfs > 1 );
  if( bad && options_.verbose_>1 ) PetscPrintf(PETSC_COMM_SELF,"[%d]%s() Bad--DONE\n",
				     comm_.mype(),__FUNCT__);
  
  return 0;
}


/* PlaneIntersect() ***********************************************
 *
 *   - solves for scale for ray(n0->vec) to intersect a plane(n1,n2,n3)
 *
 *  INPUT:
 *    - n0: coordinate of base of ray
 *    - vec: vector of ray (to be scaled with return val)
 *    - n1,n1,n3: three coordinates of the plane to hit
 *
 *  SIDE EFFECTS:
 *    - 30 flops
 *
 *  RETURN:
 *    - value to scale 'vec', zero if singular
 */
gfloat PlaneIntersect(const PromCoord &n0,const PromCoord &vec,
		      const PromCoord &n1,const PromCoord &n2,
		      const PromCoord &n3)
{
  double   a[4][5], temp, ajj, ans[3]; 
  
  a[0][0] = n2[0] - n1[0]; // normalize with n0 coord
  a[1][0] = n2[1] - n1[1]; 
  a[2][0] = n2[2] - n1[2]; 
  
  a[0][1] = n3[0] - n1[0]; 
  a[1][1] = n3[1] - n1[1]; 
  a[2][1] = n3[2] - n1[2]; 
  
  a[0][2] = -vec[0]; 
  a[1][2] = -vec[1]; 
  a[2][2] = -vec[2]; 
  
  a[0][3] = n0[0] - n1[0];  // rhs
  a[1][3] = n0[1] - n1[1]; 
  a[2][3] = n0[2] - n1[2];
  
  a[3][0] = a[3][1] = a[3][2] = a[3][3] = a[3][4] = 0.0;
  
  // forward elimination - row 2-4
  PromContext::pivot45( 0, a );
  ajj = a[0][0]; if( fabs(ajj) < 1.e-100 ) return 0.0;
  temp = a[1][0] / ajj;    
  a[1][1] -= a[0][1]*temp;
  a[1][2] -= a[0][2]*temp; 
  a[1][3] -= a[0][3]*temp; 
  
  temp = a[2][0] / ajj;    
  a[2][1] -= a[0][1]*temp;
  a[2][2] -= a[0][2]*temp; 
  a[2][3] -= a[0][3]*temp; 
  
  // forward elimination - row 3-4
  PromContext::pivot45(1, a);
  ajj = a[1][1]; if( fabs(ajj) < 1.e-100 ) return 0.0;
  temp = a[2][1] / ajj;    
  a[2][2] -= a[1][2]*temp; 
  a[2][3] -= a[1][3]*temp; 
     
  // back sub  
  ajj = a[2][2]; if( fabs(ajj) < 1.e-100 ) return 0.0;
  temp = a[2][3] / ajj;
  //a[1][3] -= a[1][2] * temp;
  //a[0][3] -= a[0][2] * temp;
  ans[2] = temp;  //copy answer
  
  //temp = a[1][3] / a[1][1];
  //a[0][3] -= a[0][1] * temp;
  //ans[1] = temp;;  //copy answer
  
  //ans[0] = a[0][3] / a[0][0];;  //copy answer 

  return ans[2];
}

/* PlaneIntersect2() ***********************************************
 *
 *   - solves for scale for ray(curr->vec) to intersect a plane(n1,n2,n3)
 *
 *  INPUT:
 *    - curr: coordinate of base of ray
 *    - vec: vector of ray (to be scaled with return val)
 *    - n1,n1,targ: three coordinates of the plane to hit
 *
 *  SIDE EFFECTS:
 *    - 28 flops
 *
 *  RETURN:
 *    - value to scale 'vec', zero if singular
 */
gfloat PlaneIntersect2( const PromCoord &curr,const PromCoord &vec,
				 const PromCoord &n1,const PromCoord &n2,
				 const PromCoord &targ)
{
  double   a[3][4], temp, ajj, ans[3]; 
  
  a[0][0] = n2[0] - n1[0]; // normalize with curr coord
  a[1][0] = n2[1] - n1[1]; 
  a[2][0] = n2[2] - n1[2]; 
   
  a[0][1] = curr[0] - n1[0]; 
  a[1][1] = curr[1] - n1[1]; 
  a[2][1] = curr[2] - n1[2]; 
  
  a[0][2] = vec[0]; 
  a[1][2] = vec[1]; 
  a[2][2] = vec[2]; 
  
  a[0][3] = targ[0] - n1[0];  // rhs
  a[1][3] = targ[1] - n1[1]; 
  a[2][3] = targ[2] - n1[2];
  
  // forward elimination - row 2-3
  pivot34( 0, a );
  ajj = a[0][0];           if( fabs(ajj) < 1.e-100 ) return 0.0;
  temp = a[1][0] / ajj;    
  a[1][1] -= a[0][1]*temp;
  a[1][2] -= a[0][2]*temp; 
  a[1][3] -= a[0][3]*temp; 

  temp = a[2][0] / ajj;    
  a[2][1] -= a[0][1]*temp;
  a[2][2] -= a[0][2]*temp; 
  a[2][3] -= a[0][3]*temp; 
  
  // forward elimination - row 3
  pivot34(1, a);
  ajj = a[1][1];           if( fabs(ajj) < 1.e-100 ) return 0.0;
  temp = a[2][1] / ajj;    
  a[2][2] -= a[1][2]*temp; 
  a[2][3] -= a[1][3]*temp; 
     
  // back sub  
  ajj = a[2][2];           if( fabs(ajj) < 1.e-100 ) return 0.0;
  temp = a[2][3] / ajj;
  a[1][3] -= a[1][2] * temp;
  //a[0][3] -= a[0][2] * temp;
  ans[2] = temp;  //copy answer
  
  ajj = a[1][1]; 
  temp = a[1][3] / ajj;
  //a[0][3] -= a[0][1] * temp;
  ans[1] = temp;  //copy answer
  
  //ans[0] = a[0][3] / a[0][0];  //copy answer 

  return (ans[2]/ans[1]);
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::GetScaleFactor"
/* PromGeoGrid::GetScaleFactor() ******************************************
 *
 *  - get scale of 'deltaVec' to move 'curr'
 *
 *  INPUT:
 *    - curr: node at center of facets 
 *    - deltaVec: vector to move 'curr' along (length is twice area of patch)
 *    - unfound: list of nodes to check for
 *    - edgeUnfound: list of nodes, from edge, to check for
 *    - edges: arrays of nodes of facets
 *    - numfacets: number of edges.
 *    - scale: scale factor. (out)
 *
 *  SIDE EFFECTS:
 *    - 
 *
 *  RETURN:
 *    - PETSc error code
 */ 
int PromGeoGrid::GetScaleFactor(PromGeoNode *curr, const PromCoord &deltaVec,
				PromList<PromStack*> *unfound, 
				PromList<PromStack*> *edgeUnfound, 
				PromGeoNode *edges[][2], int numfacets, 
				gfloat &scale )
{
  // length is area 
  scale = sqrt(deltaVec.length());
  const gfloat maxScale = .1/scale, maxtemp = .8/scale;
  scale = 0.0; // return
  
  // get scale for 'deltaVec' to move 'curr' to cover all targets
  PromList<PromStack*> *plist;
  int kk;
  for (kk = 0, plist = unfound ; kk < 2 ; kk++, plist = edgeUnfound){
    Position pos = plist->getHeadPosition();
    while ( pos ) {    
      PromGeoNode *targ = (PromGeoNode*) plist->getNext( pos );  
      //if( targ->isFixed() ) continue; // do not try move for fixed      
      // check all edges to get scale to move 'curr'
      gfloat tscale = 0.0;
      for ( int ii = 0 ; ii < numfacets ; ii++ ) {
        PromGeoNode *n1 = edges[ii][0];
        PromGeoNode *n2 = edges[ii][1];
        gfloat temp = PlaneIntersect( curr->symbGeoData_->coord,deltaVec,n1->symbGeoData_->coord,n2->symbGeoData_->coord,
				      targ->symbGeoData_->coord );
	if ( temp > tscale && temp < maxtemp ) tscale = temp; 
      }
      if ( tscale > scale ) scale = tscale;
    }
  }
  if( scale > maxScale ){
    scale = maxScale;
    if(options_.verbose_>1) PetscPrintf(gComm_,"+");
  }
  else scale *= 1.0001;
   
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::EmptyChildren"
/* PromGeoGrid::EmptyChildren() ***************************************
 *
 *  - put children to nearest unmarked neighbor
 *
 *  INPUT:
 *    curr: 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error flag
 */
int PromGeoGrid::EmptyChildren( PromGeoNode *curr )
{    

  while( !curr->isChildListEmpty() ) {
    PromGeoNode *fine = (PromGeoNode*)curr->PopChild(), *best;
    
    gfloat tt, fbest; 
    best = NULL;
    TablePos tpos = curr->symbData_->adjac.getHeadPosition();
    while ( tpos ) {
      PromGeoNode *curr2 = (PromGeoNode*)curr->symbData_->adjac.getNext(tpos);
      if(curr2->isGhost()||curr2->isMarked()||!curr2->isExterior()) continue;
      PromCoord vec = fine->symbGeoData_->coord -
	curr2->symbGeoData_->coord; tt = vec.length();
      if( !best ){ best = curr2; fbest = tt; }
      else if( tt < fbest ){ best = curr2; fbest = tt; }
    }
    if( !best ){
      if(options_.verbose_>1) PetscPrintf(PETSC_COMM_SELF,"[%d]%s expensive add\n",
				comm_.mype(),__FUNCT__);
      for( int yy = 0 ; yy < nLocalNd_ ; yy++ ) {
	PromGeoNode *curr3 = (PromGeoNode*)nodes_[yy];
	if( !curr3->isGhost() && !curr3->isMarked() && 
	    curr3->isExterior() && curr != curr3 ) {
	  PromCoord vec = fine->symbGeoData_->coord -
	    curr3->symbGeoData_->coord; tt = vec.length();
	  if( !best ){ best = curr3; fbest = tt; }
	  else if( tt < fbest ){ best = curr3; fbest = tt; }
	}
      }
    }
    if( !best ){ // no place to put, give up!!!  
      curr->addChild( fine );
      if(options_.verbose_>1) PetscPrintf(PETSC_COMM_SELF,"[%d]Q",comm_.mype());
      return 0; 
    } 
    best->addChild( fine );
  }
  
  return 0;
}
