//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prometheus.C,v 1.83 2006/10/13 19:56:29 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prometheus.C
//
#include "prometheus.hh"

static PetscLogDouble rss=0.0;
int prom_table_size = 191;

// main driver
#undef __FUNCT__    
#define __FUNCT__ "Prometheus::Prometheus" 
/* Prometheus::Prometheus *****************************************************
 * 
 */ 
Prometheus::Prometheus( const PromComm &comm, PromOptions &opts, 
			PromPerfMonitor &perf ) :
  finalized_(FALSE),
  mgType_(PROM_AGGMG),
  top_grid_(NULL),
  grid0_(NULL),
  right_(NULL),
  left_(NULL),
  solver_(NULL), 
  Comm_(comm),
  options_(opts),
  perf_mon_(perf)
{}

#undef __FUNCT__  
#define __FUNCT__ "Prometheus::~Prometheus" 
/* Prometheus::~Prometheus*************************************************
 *
 * - Prometheus destructor
 *
 */
Prometheus::~Prometheus()
{
  assert(this!=NULL);

  // delete solver
  if( right_ != NULL ) delete right_ ; 
  if( left_ != NULL && left_ != right_ ) delete left_ ; 
  if( solver_ != NULL ) delete solver_;  

  // destroy grids, after vectors but before opts(?)
  PromGrid *g = grid0_, *tg; grid0_ = NULL;
  while( (tg=g) != NULL ){ 
    g = g->nextgrid_;
    delete tg;
  }
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus::Init" 
/* Prometheus::Init ********************************************
 *
 *  - called after 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus::Init( bool init1 /*=TRUE*/ )
{  
  int ierr, i, ii; bool flag;  
  MPI_Comm comm = Comm_.MPIComm();

  if( init1 ) {
    // PETSC event log
    char titles[ LOG_TOT ][17] = { 
      "FEI: init. str. ", // LOG_FEI_INIT_STR
      "FEI: Prom setup ", // LOG_FEI_PROM_SETUP
      "FEI: solv.setup ", // LOG_FEI_MATRIX_SETUP 
      " FEI: BCs & reg.", // LOG_FEI_CR_BC
      "  FEI:reg.(athen", // LOG_PARTITION 
      " FEI: MakeRest. ", // LOG_RESTRICTION
      "FEI: SLESSolve*3", // LOG_FEI_SOLVE
      "FEI: pt.ass.(cr)", // LOG_FEI_ASSEM
      "FEI: blk.ass(Met", // LOG_METIS
      "FEI: mat. AXPY  ", // LOG_FEI_MAT_AXPY
      "FEI: resetCR    ", // LOG_FEI_RESETCR
      "Fine grid      1", // LOG_SYMB_NUMR
      "MG++(new grid) 1", // LOG_NEW_GRID, top symbolic phase  
      "G-S core        ", // LOG_FINALIZE,
      "RAP            2", // LOG_MAT_TRI_PROD
      "SLES setup    *2", // LOG_SLES_SETUP
      "Prometheus    *1", // LOG_PROM
      "FEAP formfe/C*C'", // LOG_FEAP
      "Coarse Grid Solv", // LOG_ASSEMBLE
      "Grid coarsen   1", // LOG_COARSEN
      "symb_i_feap     ", // LOG_SYMB_INIT_FEAP
      "New Coarse     1", // LOG_NEW_COARSE
      "Tessellate      ", // LOG_TRIANGULATE
      "Fix mesh        ", // LOG_FIXMESH
      "Exact adjacent 1", // LOG_EXACTADJ
      "SmoothP        2", // LOG_MAPS
      "CG Est. lamb.1 2", // LOG_MIS0
      "AP_0           2", // LOG_MIS1
      "G-S setup       ", // LOG_MIS2 
      "Gauss-Seidel    ", // LOG_MIS3
      "G-S Infs (recv) ", // LOG_MIS4
      "G-S Sups        ", // LOG_MIS5
      " RAP core      2", // LOG_MIS6
      " RAP my assem. 2", // LOG_MIS7
      "Repartition    1", // LOG_REPART
      "MDot  (GMRES)   ", // LOG_AUX1
      "Symm mat. (Aux2)", // LOG_AUX2
      "Sq. graph (Aux3)"};// LOG_AUX3
    for( ii = 0 ; ii < LOG_TOT ; ii++ ) {
      ierr = perf_mon_.EventRegister(&prom_logs_[ii], titles[ii],0); 
      CHKERRQ(ierr);
    }
#if !defined(PROM_NO_STAGES)
    perf_mon_.StageRegister(&prom_stages_[0],"Part/FEAP/set"); // 0
    perf_mon_.StageRegister(&prom_stages_[1],"MG Level 0   "); // 1
    perf_mon_.StageRegister(&prom_stages_[2],"MG Level 1   "); // 2
    perf_mon_.StageRegister(&prom_stages_[3],"MG Level 2   "); // 3
    perf_mon_.StageRegister(&prom_stages_[4],"MG Level 3   "); // 4
    perf_mon_.StageRegister(&prom_stages_[5],"Prom Test    "); // 5
#endif
  }
  // creates options database and perf monitor (may not be used)
  if(PROM_BSZ < Comm_.np()) SETERRQ(1,"PROM_BSZ < Comm_.np()");
  ierr = options_.HasOptionName( "-out_matrix", &flag );      CHKERRQ(ierr);
  if( flag ) options_.matrix_ = TRUE;
  else options_.matrix_ = FALSE;
  ierr = options_.HasOptionName( "-out_files", &flag );       CHKERRQ(ierr);
  if ( flag ) options_.files_  = TRUE; 
  else  options_.files_ = FALSE;

  // geometric/AGG MG
  ierr = options_.HasOptionName( "-prometheus_geomg", &flag );CHKERRQ(ierr);
  if( flag ) mgType_ = PROM_GEOMG;
  else{
    ierr = options_.GetOptionInt("-aggmg_smooths", &ii, &flag); CHKERRQ(ierr);
    if( flag && ii < 0 )  mgType_ = PROM_GEOMG;
    else mgType_ = PROM_AGGMG; // aggmg is default
  }

  // verbose
  ierr = options_.GetOptionInt("-out_verbose", &i, &flag);   CHKERRQ(ierr);
  if( flag ){
    options_.verbose_ = i;
    if( options_.verbose_>2 ) {
      i = sizeof(PromTable<int>); ii = sizeof(PromDTet);int psz=sizeof(int*);
      PetscPrintf(comm, "\tSize of (bytes):\n\t\tPromAggNode = %d\n\t\tPromGeoNode = %d\n\t\tPromDTet = %d\n\t\tPromList<int> = %d\n\t\tPromTable<int> = %d\n\t\tPromContext = %d\n\t\tpointer = %d\n\t\tPromHex = %d\n\t\tPromElement = %d\n\t\tPromMaterial = %d\n\t\tPromStack = %d\n",
		  sizeof(PromAggNode),sizeof(PromGeoNode), ii,
		  sizeof(PromList<int>),i,sizeof(PromContext),psz,
		  sizeof(PromHex), sizeof(PromElement), sizeof(PromMaterial),
		  sizeof(PromStack) );
    }
  }
  else options_.verbose_ = 0;
  // output 
  if( options_.verbose_ > 1 && init1 ) {
    PetscMemoryGetCurrentUsage( &rss );
    PetscPrintf(comm,"[%d]ASSERT on: %d pc,%d Mb,|bits|=%d,|d|=%d,|f|=%d,|i|=%d,|p|=%d PROM_BSZ=%d\n",
		Comm_.mype(),Comm_.np(),(int)(rss/1.e6+.5),
		sizeof(struct bitsTAG),sizeof(double),
		sizeof(float),sizeof(int),sizeof(int*),PROM_BSZ);
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix_base::PromMatrix_base" 
/* PromMatrix_base::PromMatrix_base **************************************
 *
 */
PromMatrix_base::PromMatrix_base( const PromMap &map ) : 
  map_(map), numRecv_(0), pRecvDone_(FALSE), recvSzs_(0), recvParts_(0), 
  proci_table_gidi_tablej_(41), proci_table_gidi_tablej_i_(41),gid_ndf_(NULL),
  DRhoA_(-1.0), dirty_data_(FALSE), indefinite_(0)
{
}

#undef __FUNCT__  
#define __FUNCT__ "PromMap_base::PromMap_base" 
/* PromMap_base::PromMap_base **************************************
 *
 */
PromMap_base::PromMap_base( int nloc, int nloceq, const PromComm_base &comm ) 
  : proc_gnode_(NULL), proc_geq_(NULL), comm_(comm)
{
  MPI_Comm mpi = comm(); assert(mpi!=MPI_COMM_NULL); int np = comm.np();
  PetscMalloc( (np+2)*sizeof(int), &proc_gnode_ ); assert(proc_gnode_!=NULL);
  PetscMalloc( (np+2)*sizeof(int), &proc_geq_ ); assert(proc_geq_!=NULL);
  ResetMap(nloc,nloceq);
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix_base::DeleteProctable"
/* PromMatrix_base::DeleteProctable ************************************
 *
 *  - delete "proci_table_gidi_tablej"
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code 
 */ 
int PromMatrix_base::DeleteProctable()
{ 
  PromTable<PromTable<double*>*> *rowi_table_rowj;
  while( !proci_table_gidi_tablej_.isEmpty() ) { 
    rowi_table_rowj = proci_table_gidi_tablej_.removeHead();
    while( !rowi_table_rowj->isEmpty() ) { 
      PromTable<double*> *rowj_vals = rowi_table_rowj->removeHead();
      while( !rowj_vals->isEmpty() ) { 
	double *vals = rowj_vals->removeHead();
	PetscFree( vals );
      } // j
      delete rowj_vals; 
    } // i
    delete rowi_table_rowj; 
  }
  while( !proci_table_gidi_tablej_i_.isEmpty() ) { 
    rowi_table_rowj = proci_table_gidi_tablej_i_.removeHead();
    while( !rowi_table_rowj->isEmpty() ) { 
      PromTable<double*> *rowj_vals = rowi_table_rowj->removeHead();
      while( !rowj_vals->isEmpty() ) { 
	double *vals = rowj_vals->removeHead();
	PetscFree( vals );
      } // j
      delete rowj_vals; 
    } // i
    delete rowi_table_rowj; 
  }
  if( numRecv_ ){
    if( recvSzs_ != NULL ) PetscFree( recvSzs_ ); 
    if( recvParts_ != NULL ) PetscFree( recvParts_ );
  }
  numRecv_ = 0;
  recvSzs_ = recvParts_ = NULL;
  pRecvDone_ = FALSE;
  if( gid_ndf_ != NULL ) { delete gid_ndf_; gid_ndf_ = NULL; }

  return 0;
}

#if defined(PROM_USE_PETSC_PERF_MON)

int PromPerfMonitor::Print( MPI_Comm comm, char *fname /*=NULL*/, 
			    bool /*=FALSE*/ ) const
{
  int ierr;  
  ierr = PetscLogPrintSummary( comm, fname ); CHKERRQ(ierr);

  return 0;
}

#else
#undef __FUNCT__  
#define __FUNCT__ "PromPerfMonitor::Print" 
/* PromPerfMonitor::Print*************************************************
 *
 * - Prometheus destructor
 *
 */
int PromPerfMonitor::Print( MPI_Comm comm, char *fname /*=NULL*/, 
			    bool print /*=FALSE*/ ) const
{
  assert(this!=NULL);
  int  ierr,rank; const int MFLOP=0,TIME=1,MFLOP_S=2,COUNT=3;
  FILE *file = NULL;
  
  if( fname == NULL ) file = stdout;
  else {
    file = fopen( fname, "w" );
    if( !file ) SETERRQ(1,fname);
  } 
  if( comm != MPI_COMM_NULL ) {
    ierr = MPI_Comm_rank(comm,&rank); CHKERRQ(ierr);
  }
  else rank = -1;
  if( rank == 0 ){
    fprintf(file,
"\t%g Mflops total.  %d Events\n",flops_*1.e-6,size_);
    fprintf(file,
"                     Count            Time              MegaFlops/sec\n");
    fprintf(file,
"   Event        :  Max  Ratio      Max    Ratio      Max    Ratio    Total\n");
  }
  double t1,t2,sb[4];
  for( int st = 0 ; st < 10 ; st++ ) { 
    if( rank == 0 ) fprintf(file,"%d:        Stage: %s\n",st,stagenames_[st]);
    for( int ii = 0 ; ii < size_ ; ii++ ) {
      const PromPerfEvent * const ev = &eventTable_[ii];
      if( ev->bits_[0].reged == 0 ) continue;
      t1 = ev->flops_[st]; t2 = ev->time_[st];
      sb[MFLOP] = t1*1.e-6; sb[TIME] = t2; 
      sb[COUNT] = (double)ev->bits_[st].count;
      if( t2 > 0.0 ) sb[MFLOP_S] = ev->flops_[st]*1.e-6/t2;
      else sb[MFLOP_S] = 0.0;
      double rb_sum[4],rb_max[4],rb_min[4];
      MPI_Allreduce(sb,rb_sum,4,MPI_DOUBLE,MPI_SUM,comm);
      if( rb_sum[COUNT] == 0.0 ) continue;
      MPI_Allreduce(sb,rb_max,4,MPI_DOUBLE,MPI_MAX,comm);
      MPI_Allreduce(sb,rb_min,4,MPI_DOUBLE,MPI_MIN,comm);
      assert(rb_sum[TIME] >= 0.0); assert(rb_sum[TIME] >= rb_max[TIME]); 
      if( rank == 0 || print ) {
	t1 = (rb_min[COUNT]>0.0) ? rb_max[COUNT]/rb_min[COUNT] : 0.0;
	t2 = (rb_min[MFLOP_S]>0.) ? rb_max[MFLOP_S]/rb_min[MFLOP_S] : 1.;
	fprintf(file,"%16s:%5d %5.1g | %7.3e %5.1g | %7.3e %5.1g %7.3e\n",
		ev->name_, (int)rb_max[COUNT], t1,
		rb_max[TIME], 
		(rb_min[TIME]>0.0) ? rb_max[TIME]/rb_min[TIME] : 1.,
		rb_max[MFLOP_S], t2,
		(rb_max[MFLOP_S]>0.0) ? rb_sum[MFLOP]/rb_max[MFLOP_S] : 0.0 );
      }
    }
  }
  if( file != stdout ) fclose(file);

  return 0;
}
#endif

#undef __FUNCT__
#define __FUNCT__ "Prometheus::NewLevel"
/* Prometheus::NewLevel *************************************************
 *
 *
 *  INPUT: 
 *   nnodes_top: (in/out) 'is last' flag in, number of nodes on coarse grid out
 *
 *  SIDE EFFECTS:
 */ 
int Prometheus::NewLevel( int tlevel )
{ 
  int ierr,xx;
  const int np=Comm_.np(),mype=Comm_.mype(),verb=(options_.verbose_>1);
  MPI_Comm wcomm = Comm_.MPIComm();
  bool flag;
  const bool islast = (tlevel == 0);

  // reality checks
  assert( top_grid_ != NULL );
  if( finalized_ ) {  
    if(verb)PetscPrintf( wcomm,"[%d]%s MG++ called after finalized\n",
			 mype, __FUNCT__);
    return -1; 
  }

  // HELPERS
  PromGrid *lastGrid = top_grid_;
  if( !lastGrid ) lastGrid = grid0_; // fine grid
  const int lev1 = lastGrid->getLevel(), lev2 = lev1+1;
  
  // SETUP FINE GRID
  if( lastGrid == grid0_ ) { // fine grid setup
    if(verb)PetscPrintf(wcomm,"[%d]%s Init fine grid\n",mype,__FUNCT__);
    ierr = perf_mon_.EventBegin(prom_logs_[LOG_PROM]);      CHKERRQ(ierr); 
    ierr = perf_mon_.EventBegin(prom_logs_[LOG_SYMB_NUMR]); CHKERRQ(ierr);
    // assemble fine grid (not timed) -- side effect
    assert( grid0_->stiffness_ != NULL );
    ierr = SetMatrixWithNewData( grid0_->stiffness_ );      CHKERRQ(ierr);
    ierr = perf_mon_.EventEnd(prom_logs_[LOG_SYMB_NUMR] );  CHKERRQ(ierr);
    ierr = perf_mon_.EventEnd(prom_logs_[LOG_PROM]);        CHKERRQ(ierr);
  }

  // SEE IF NO MORE GRIDS - global 
  options_.GetOptionInt( "-prometheus_top_grid_limit", &xx, &flag );
  if( flag && xx < 10 ) xx = 10;
  const int topMaxEq = (flag) ? xx : 2500;
  if( allDone(lastGrid,topMaxEq) ) return 0; 
  if(verb)PetscPrintf(wcomm,"[%d]%s make new grid\n",mype,__FUNCT__);

  // COARSEN 
  ierr = perf_mon_.EventBegin(prom_logs_[LOG_PROM]); CHKERRQ(ierr);
  ierr = perf_mon_.EventBegin(prom_logs_[LOG_COARSEN]); CHKERRQ(ierr);
  // smoothed aggregation coarsening
  if( lastGrid->stiffness_ == NULL ){
    SETERRQ(1,"No stiffness matrix - call assemble before prom_new_level");
  }
  if( lastGrid->isActive() ) {
    if(verb)PetscPrintf(wcomm,"[%d]%s call Coarsen()\n",mype,__FUNCT__);
    ierr = lastGrid->Coarsen(); CHKERRQ(ierr);
    if( lastGrid->prevgrid_ && 
	lastGrid->prevgrid_->factor_ != lastGrid->factor_ ) {
      ierr = lastGrid->SendStatesToSlaves();  CHKERRQ(ierr); 
    }
  }
  else if( lastGrid->prevgrid_ && lastGrid->prevgrid_->isActive() && 
	   lastGrid->prevgrid_->factor_ != lastGrid->factor_ ){
    ierr = lastGrid->ReceiveStates();      CHKERRQ(ierr);
  }
  ierr = perf_mon_.EventEnd(prom_logs_[LOG_COARSEN]); CHKERRQ(ierr);
  if(verb)PetscPrintf(wcomm,"[%d]%s(%d)Coarsen done\n",mype,__FUNCT__,lev2);

  PromGrid *newGrid = NULL;
  ierr = perf_mon_.EventBegin(prom_logs_[LOG_NEW_COARSE]); CHKERRQ(ierr);      
  ierr = MakeNewGrid(islast, topMaxEq, lastGrid, newGrid); CHKERRQ(ierr);
  ierr = perf_mon_.EventEnd( prom_logs_[LOG_NEW_COARSE] ); CHKERRQ(ierr);
  if( newGrid == NULL ){    
    ierr = perf_mon_.EventEnd( prom_logs_[LOG_PROM] ); CHKERRQ(ierr);
    if(options_.verbose_>0)
      PetscPrintf(wcomm,"[%d]%s no new grid\n",mype,__FUNCT__);
    return 0; // 
  }

  // CONSTRUCT COARSE GRID SPACES
  ierr = perf_mon_.EventBegin( prom_logs_[LOG_NEW_GRID] ); CHKERRQ(ierr); 
  if( lastGrid->isActive() ) {
    if(verb)PetscPrintf( wcomm, "[%d]%s CreateCoarseGridSpace(%d)\n",
			 mype, __FUNCT__, lev2 );
    ierr = newGrid->CreateCoarseGridSpace( lastGrid ); CHKERRQ(ierr);
  }
  ierr = perf_mon_.EventEnd( prom_logs_[LOG_NEW_GRID] ); CHKERRQ(ierr);

  // split data structures and clean up (before repartitioning)
  for( xx = 0; xx < newGrid->nNodes_ ; xx++ ){
    PromNode *curr = newGrid->nodes_[xx];
    assert( !curr->symbData_->parent || curr->getState() == PROM_DELETED);
    assert( !curr->isMarked() );
    curr->symbData_->child = NULL; // split here
  } 

  // MAKE NUMERIC
  if(verb)PetscPrintf( wcomm,"[%d]%s MakeNumeric(%d)...\n", 
		       mype,__FUNCT__, lev1 );
  ierr = lastGrid->PrepForSolve();    CHKERRQ(ierr);
  top_grid_ = newGrid;

  // Repartiton lastgrid
  ierr = options_.HasOptionName("-prometheus_repartition",&flag);
  CHKERRQ(ierr);
  if( flag && newGrid->factor_ != np ){
    ierr = options_.HasOptionName("-prometheus_use_deterministic",&flag);
    if( !flag ){
      if(verb)PetscPrintf( wcomm, "[%d]%s Repartition(%d)\n",
			   mype,__FUNCT__,lev2);
      ierr = newGrid->Repartition( lastGrid ); CHKERRQ(ierr); 
    }
  }

  { // deal with blocks
    int nblocks = 0, neqTot = newGrid->nglobalEq();
    //if( newGrid->factor_ == np ) nblocks = 1; else
    if ( lastGrid->nBlocks_ > 0 ) {
      int totNNB = nBlocksWithGlobalDOF( neqTot ), neq =  newGrid->nlocalEq();
      if( !newGrid->isActive() ) nblocks = 1;
      else nblocks = (int)((float)totNNB*(float)neq/(float)neqTot);
      if( totNNB > 0 && nblocks < 1 ) nblocks = 1;
    }
    newGrid->nBlocks_ = nblocks; // local n blocks
  }
  ierr = perf_mon_.EventEnd( prom_logs_[LOG_PROM] ); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus::ConvertToSymm"
/* Prometheus::ConvertToSymm *************************************************
 *
 *
 *  INPUT: 
 *   A:
 *   mg:
 *
 *  SIDE EFFECTS:
 */ 
int Prometheus::ConvertToSymm(PromMatrix *const AA, const PromOptions &opts ) const
{
  bool flag; int vb=opts.verbose_; 
  char str[32]; int ierr,mype=AA->map_.comm_.mype(); 
  MPI_Comm comm = AA->map_.comm_(); 
  ierr = opts.GetOptionString("-mg_levels_pc_type",str,32,&flag); 
  CHKERRQ(ierr);
  if( flag && (strcmp(str,"gs")==0 || strcmp(str,"gs_asm")==0) ) return 0;
  return 0; // disabled!!!
  if( !flag ) {
    ierr = opts.GetOptionString("-pc_type",str,32,&flag); CHKERRQ(ierr);
    if(flag && (strcmp(str,"gs")==0 || strcmp(str,"gs_asm")==0)) return 0;
  }
  ierr = opts.HasOptionName("-pc_symmetric",&flag); CHKERRQ(ierr);
  if( flag ){
    if(vb>1)PetscPrintf(comm,"[%d]%s Call ConvertToSymm (N=%d)\n",mype,__FUNCT__,AA->getM());
    ierr = perf_mon_.EventBegin( prom_logs_[LOG_AUX2] ); CHKERRQ(ierr);
    ierr = AA->ConvertToSymm(); CHKERRQ(ierr); 
    ierr = perf_mon_.EventEnd( prom_logs_[LOG_AUX2] ); CHKERRQ(ierr);
    if(vb>2)PetscPrintf(comm,"[%d]%s ConvertToSymm done\n",mype,__FUNCT__);
  }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus::allDone"
/* Prometheus::allDone ************************************************
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - do not make another grid
 */  
bool Prometheus::allDone( PromGrid *lastGrid, const int topMaxEq ) const
{
  int xx,num,np=Comm_.np(),mype=Comm_.mype(),verb=(options_.verbose_>1);
  const int oldv = lastGrid->nglobalEq(), lastfact = lastGrid->factor_;
  MPI_Comm comm = Comm_.MPIComm();

  if( lastGrid->isActive() ) { 
    if( !lastGrid->isBase() ) {
      if( oldv<topMaxEq || (lastfact == np && np > 1) ){ 
	if(verb)
	  PetscPrintf(comm,
		      "\t[%d]%s TOP small (N<%d)=%d (lastfactact==np) = %d\n",
		      mype,__FUNCT__, topMaxEq, oldv<topMaxEq,
		      (lastfact==np && np>1) );
	num = 0;
      }
      else num = 1;
    }
    else num = 1;
    // comm
    if( lastfact != 1 ){
      xx = lastfact - 1;
      do{ 
	MPI_Send( &num, 1, MPI_INT, mype + xx, 0, comm );    
      }while( --xx );
    }
  } 
  else {
    // comm
    MPI_Status       status; 
    xx = mype%lastfact; assert(xx>0);
    MPI_Recv( &num, 1, MPI_INT, mype - xx, 0, comm, &status );
  }
  return (num==0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus::MakeNewGrid"
/* Prometheus::MakeNewGrid ************************************************
 *
 *  INPUT:
 *    - islast: is this the last grid (not necc.)
 *    - topMaxEq: if the next grid has less, then this 'islast'
 *    - lastGrid:
 *    - newGrid: (out) 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - error code
 */  
int Prometheus::MakeNewGrid( bool islast, const int topMaxEq, 
			     PromGrid *const lastGrid, 
			     PromGrid *& newGrid ) const
{
  int ierr,neqTot,minloc,nloc,nnTot,neq,nextf,nnodes,num,xx,ndf;
  const int np=Comm_.np(),mype=Comm_.mype(),verb=(options_.verbose_>1);
  const int lev1=lastGrid->getLevel(),lev2=lev1+1,lf=lastGrid->factor_;
  bool flag, tit;
  MPI_Comm comm = Comm_.MPIComm();

  PetscFunctionBegin;
  if( lastGrid->isActive() ) {
    ierr = lastGrid->NumSelected( nloc, nnodes, neq );     CHKERRQ(ierr);
  }
  else nnodes = nloc = neq = 1000000;
  if(verb)PetscPrintf(verb > 1 ? PETSC_COMM_SELF : comm,
		      "[%d]%s Make level %d, %d local selected\n",
		      mype,__FUNCT__,lev2,nloc);
  int min3[3], ms[3]={ nloc, nnodes, neq }; 
  MPI_Allreduce( ms, min3, 3, MPI_INT, MPI_MIN, comm );
  minloc = min3[0];
  if(!lastGrid->isActive()) neq = nnodes = nloc = ms[0] = ms[1] = ms[2] = 0;
  else{ assert(ms[0]==nloc); ms[1]=lastGrid->nBlocks_; assert(ms[2]==neq);}
  MPI_Allreduce( ms, min3, 3, MPI_INT, MPI_SUM, comm );
  nnTot = min3[0]; neqTot = min3[2]; 
  if( nnTot == lastGrid->nglobal() ) return 0; // no reduction
  // before new grid, count nodes, compute num active procs  
  ierr = options_.HasOptionName("-prometheus_use_parallel_top_solver",&tit);
  CHKERRQ(ierr);
  if( islast && !tit ) nextf = np; 
  else {
    ierr = options_.HasOptionName("-prometheus_flat", &flag);  CHKERRQ(ierr);
    if( flag ) nextf = lf; 
    else {
      ierr = GetNextFactor( options_, np, lf, neqTot, &num ); CHKERRQ(ierr);
      if(verb)PetscPrintf(comm,"[%d]%s GetNextFactor(%d)=%d Neq=%d\n",
			  mype, __FUNCT__,lev2,num,neqTot);
      nextf = (neqTot < topMaxEq && !tit ) ? np : num;
      if( nextf < lf ) nextf = lf; // ???
      assert(nextf >= lf && nextf <= np);
    }
  }
  // odd case
  if( lf == nextf && nextf != 1 ){ 
    for( xx = 2 ; lf == nextf ; xx++ ) {
      if( np%(lf*xx) == 0 ) nextf = lf * xx;
    }
  }

  if(verb)
    PetscPrintf(comm,
		"\t[%d]%s minloc=%d,%d gnodes,last loc(%d,...,%d)nextf=%d (lastG is %sconst ndf)\n",
		mype,__FUNCT__, minloc, nnTot,
		lastGrid->proc_gnode()[1] - lastGrid->proc_gnode()[0],
		lastGrid->nglobal() - lastGrid->proc_gnode()[np-lf],
		nextf,lastGrid->isConstNDF() ? "" : "NON-");
  ndf = 0;
  if( lastGrid->isConstNDF() ) {
    for( num=0, xx=-1 ; num < lastGrid->nLocalNd_ && xx <= 0 ; num++ ) { 
      xx = lastGrid->nodes_[num]->getNextNDF();
    }
    MPI_Allreduce(&xx,&ndf,1,MPI_INT,MPI_MAX,comm);assert(xx<=0 || xx==ndf);
    assert(xx==-1 || xx==ndf);
  }

  // create coarse grid 
  if( lf == nextf ){ // no split
    assert(lastGrid->isActive()); // would be wierd to not reduce after start
    // normal create new coarser grid (from selected flags on finer grid)
    if(verb)PetscPrintf(comm,"[%d]%s Normal New Grid(%d) %s\n",mype,__FUNCT__,
			lev2,lastGrid->isConstNDF()?"const NDF":"variable NDF");
    if( lastGrid->isConstNDF() ) {
      if( mgType_ == PROM_AGGMG ) { 
	newGrid = new PromAggGrid( options_, perf_mon_,
				   dynamic_cast<PromAggGrid*>(lastGrid), nloc,
				   nnodes, Comm_, ndf );
      }
      else if( mgType_ == PROM_GEOMG ) {
	newGrid = new PromGeoGrid( options_, perf_mon_,
				   dynamic_cast<PromGeoGrid*>(lastGrid), nloc,
				   nnodes, Comm_, ndf );
      }
      else assert(0);
    }
    else { // variable blocks row
      int *lid_ndof = NULL, ii; // general dofs!
      ierr = PetscMalloc((nloc+1)*sizeof(int), &lid_ndof);     CHKERRQ(ierr);
      for( xx = ii = 0 ; xx < lastGrid->nLocalNd_; xx++ ) {
	PromNode* curr = lastGrid->nodes_[xx]; 
	if( curr->getState() == PROM_SELECTED ) {
	  ndf = curr->getNextNDF(); lid_ndof[ii++] = ndf; assert(ndf>0);
	}
      }
      assert(ii==nloc);
      if( mgType_ == PROM_AGGMG ) { 
	newGrid = new PromAggGrid( options_, perf_mon_,
				   dynamic_cast<PromAggGrid*>(lastGrid), nloc,
				   nnodes, Comm_, lid_ndof );
      }
      else if( mgType_ == PROM_GEOMG ) {
	SETERRQ(1,"GEOMG w/ non-const ndf\n");
      }
      else assert(0);
      ierr = PetscFree(lid_ndof);  CHKERRQ(ierr);
    }
    CHKNEW(newGrid,1); 
  }
  else{
    // split
    if( lastGrid->isActive() ) {
      assert(nextf <= np);
      if(verb)PetscPrintf(comm,"[%d]%s Split top(%d): %d loc. nd.\n",
			  mype,__FUNCT__, lev2, minloc);
      ierr = MakeSplitCoarseGrid(lastGrid, nloc, nnodes, &newGrid, nextf);
      CHKERRQ(ierr);
    }
    else{
      // dummy grid
      if( lastGrid->isConstNDF() ) {
	if( mgType_ == PROM_AGGMG ) {
	  newGrid = new PromAggGrid(options_, perf_mon_,
				    dynamic_cast<PromAggGrid*>(lastGrid),
				    nextf, Comm_, ndf );
	}
	else if( mgType_ == PROM_GEOMG ) {
	  newGrid = new PromGeoGrid( options_, perf_mon_,
				     dynamic_cast<PromGeoGrid*>(lastGrid),
				     nextf, Comm_, lastGrid->constNDFSize() );
	}
      }
      else {
	if( mgType_ == PROM_AGGMG ) { 
	  newGrid = new PromAggGrid( options_, perf_mon_,
				     dynamic_cast<PromAggGrid*>(lastGrid),
				     nextf, Comm_ );
	}
	else if( mgType_ == PROM_GEOMG ) {
	  SETERRQ(1,"GEOMG w/ non-const ndf\n");
	}
      }
    }
    assert(newGrid->factor_==nextf);
  } 
  PetscFunctionReturn(0);
}

#undef __FUNCT__  
#define __FUNCT__ "Prometheus::MakeSplitCoarseGrid"
/* Prometheus::MakeSplitCoarseGrid ******************************
 *
 *  - Set up for construction of "split" procs
 *
 *  INPUT:
 *    - lastGrid: fine grid
 *    - nloc: new coarse grid: local nodes in this proc
 *    - nnodes: new coarse grid local nodes in this group
 *    - newGrid: new coarse grid (out)
 *    - factor
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int Prometheus::MakeSplitCoarseGrid( PromGrid *lastGrid, const int nloc, 
				     const int nnodes, PromGrid **newGrid,
				     const int factor ) const
{ 
  PromNode *fine; assert(lastGrid->isActive());
  MPI_Comm lcomm = lastGrid->gComm_;
  const int lfact=lastGrid->factor_,np=Comm_.np(),mype=Comm_.mype();
  const int verb=(options_.verbose_>1);
  int ii,xx,ierr,pid,*pb,kk,proc,*buff,tt,ttt,neq,ndf;
  PromGlobalID *fdep;Position pos;
#define ISLOCAL(p) (p>=master&&p<endp)
#define ISSAMEGROUP(p,p2) ((p/factor) == (p2/factor))  
  if(verb)PetscPrintf(lcomm,"[%d]%s factor=%d lastfactor=%d\n",
		      mype,__FUNCT__,factor,lfact);
  assert(factor%lfact == 0); assert(factor>lfact);
  const int master = (mype/factor)*factor; assert(master%lfact==0);
  const int endp = (master+factor > np) ? np : master+factor; 
  const int tag1=lastGrid->getNewTag(),tag2=lastGrid->getNewTag(),tag3=lastGrid->getNewTag();
  MPI_Status status; 
  MPI_Request *lreqs = NULL, *greqs = NULL; 
  int *nldata = NULL, *ngdata = NULL, ndof = -1;
  PromMPINode  mpinode;

  // mark extended fine nodes that others will do, count ghosts
  for( tt = xx = neq = 0 ; xx < lastGrid->nNodes_ ; xx++ ){
    fine = lastGrid->nodes_[xx];    fine->mark(FALSE); 
    if( fine->getState() != PROM_SELECTED ) continue; 
    if( xx < lastGrid->nLocalNd_ ) {
      ndf = fine->getNextNDF(); assert(ndf>0);
      neq += ndf; // local eqs.
    }
    // ghosts
    if( !fine->isExtended() ) continue; 
    //
    kk = 1;
    pos=fine->symbData_->dependents->getHeadPosition(); assert(pos!=NULL);
    while( pos && kk == 1 ) { // skip ones that other will do
      fdep = fine->symbData_->dependents->getNext( pos );
      proc = fdep->proc(); 
      if( ISSAMEGROUP(proc, mype) && proc > mype ) kk = 0; 
      assert(proc != mype);
    }

    if( kk == 0 ) fine->mark(); // extended node that other (p>mype) will do
    else{
      if( xx >= lastGrid->nLocalNd_ ) { // ghost that I have -- real ghost
	// need to send this as it a real ghost
	if( !ISLOCAL(fine->symbData_->ghostID.proc()) ) tt++; 
      }
    } 
  }
  const int nghost = tt; 
  // send fine counts
  int maxsizes[2]={nloc,nghost},totsz[3]={nloc,nghost,neq},rsizes[3];
  for( xx = master, kk = 0 ; xx < endp ; xx += lfact, kk++ ) {
    if( xx != mype ) MPI_Send(totsz, 3, MPI_INT, xx/lfact, tag1, lcomm);
  }
  const int nlp = kk; assert(nlp > 0);
  // receive fine counts
  for( xx = master, tt = 0 ; xx < endp ; xx += lfact ) {
    if(xx != mype) {
      MPI_Recv(rsizes,3,MPI_INT,xx/lfact,tag1,lcomm,&status);
      if( rsizes[0] > maxsizes[0] ) maxsizes[0] = rsizes[0];
      if( rsizes[1] > maxsizes[1] ) maxsizes[1] = rsizes[1];
      totsz[0] += rsizes[0]; totsz[1] += rsizes[1]; totsz[2] += rsizes[2];
      if( xx < mype ) tt += rsizes[1]; // number of ghosts before me 
    } 
  }
  PromMPINode::GetWrite_size(ii); 
  const int nGhostBeforeMe = tt; size_t mpi_node_sizeof = sizeof(int)*ii;
  const int ngbytes = maxsizes[1]*mpi_node_sizeof,ngi=ngbytes/sizeof(int); 
  const int nlbytes = maxsizes[0]*mpi_node_sizeof,nli=nlbytes/sizeof(int);
  if( nloc ) {
    ierr = PetscMalloc( nloc*mpi_node_sizeof,&nldata );     CHKERRQ(ierr);
  }
  if( nghost > 0 ) {
    ierr = PetscMalloc( nghost*mpi_node_sizeof,&ngdata );  CHKERRQ(ierr);
  }
  else ngdata = NULL;
  const int ntlocpart = totsz[0], nLocPartEq=totsz[2];

  // map type --
  if( lastGrid->isConstNDF() ) {
    for( ii=0, xx=-1 ; ii < lastGrid->nLocalNd_ && xx <= 0 ; ii++ ) { 
      xx = lastGrid->nodes_[ii]->getNextNDF();
    }
    MPI_Allreduce(&xx,&ndof,1,MPI_INT,MPI_MAX,lcomm);
    assert(xx==-1 || xx==ndof);
    tt = 1;
  }
  else {
    tt = 0; // check for const now -- TODO!!!
  }
  const bool iscondndf = (tt!=0);

  // copy fine data - all local
  for( pid = xx = 0, pb = nldata ; xx < lastGrid->nLocalNd_ ; xx++ ) {
    fine = lastGrid->nodes_[xx];
    if( fine->getState() != PROM_SELECTED ) continue;
    ierr = fine->LiftCopyIn( &mpinode );  CHKERRQ(ierr);
    mpinode.index_ = pid;
    assert((size_t)pb%sizeof(gfloat) == 0);
    mpinode.Write( &pb );
    pid++;
  }
  assert((size_t)(pb-nldata) == nloc*mpi_node_sizeof/sizeof(int));
  const int loclszi = pb-nldata;  assert(pid==nloc); assert(loclszi<=nlbytes);
  
  // copy fine data - my ghosts
  for( xx = lastGrid->nLocalNd_, pb = ngdata, ttt = totsz[0] + nGhostBeforeMe; 
       xx < lastGrid->nNodes_ ; xx++ ){
    fine = lastGrid->nodes_[xx];
    if( fine->getState() != PROM_SELECTED || fine->isMarked() ) continue;
    // 
    if( !ISLOCAL( fine->symbData_->ghostID.proc() ) ) {
      ierr = fine->LiftCopyIn( &mpinode );  CHKERRQ(ierr);
      mpinode.index_ = ttt++;
      ierr = mpinode.Write( &pb );  CHKERRQ(ierr);  // my real ghost
    }
  } 
  assert((size_t)(pb-ngdata)==nghost*(mpi_node_sizeof/sizeof(int)));
  
  // send fine ghost nodes
  tt = (pb-ngdata)/(mpi_node_sizeof/sizeof(int));
  const int locgszi = pb-ngdata; 
  assert((locgszi%(mpi_node_sizeof/sizeof(int)))==0);
  ierr = PetscMalloc(nlp*sizeof(MPI_Request),&greqs); CHKERRQ(ierr);
  for( xx = master, tt = 0 ; xx < endp ; xx += lfact, tt++ ) {
    MPI_Isend(ngdata, locgszi, MPI_INT, xx/lfact, tag2, lcomm, &greqs[tt]);
  }
  assert(tt==nlp); 

  // receive ghosts
  PromList<PromMPINode*> ghostlist;
  PromTable<int> gh_fgid1__flid1_proc1( maxsizes[1]*factor + 17 );
  if( ngbytes > 0 ){ ierr = PetscMalloc(ngbytes,&buff); CHKERRQ(ierr); }
  else buff = NULL;
  for( xx = master ; xx < endp ; xx += lfact ) {
    MPI_Recv( buff, ngi, MPI_INT, xx/lfact, tag2, lcomm, &status);
    MPI_Get_count( &status, MPI_INT, &kk ); 
    assert(kk%(mpi_node_sizeof/sizeof(int))==0);
    for( tt = 0, pb = buff ; pb-buff != kk ; tt++ ) {
      assert( (size_t)pb%sizeof(gfloat) == 0);
      ierr = mpinode.Read( &pb ); CHKERRQ(ierr); 
      for( ii = 0; ii < mpinode.getNDF() ; ii++ ) mpinode.setFixed(ii,FALSE);
      assert( !ISLOCAL(mpinode.ghostID_.proc(ierr)) );
      PromMPINode *np = new PromMPINode(mpinode); 
      CHKNEW(np,sizeof(PromMPINode));
      // -proc for ghost that I am not resp. for
      ii = gh_fgid1__flid1_proc1.add( -(xx+1),np->index_.global()+1 );
      assert(ii);
      ghostlist.addTail( np );
    } 
    assert(kk == pb - buff); 
  } 
  if(buff){ PetscFree(buff); buff = NULL;}
  // clean up
  for( xx = master ; xx < endp ; xx += lfact ){
    MPI_Waitany( nlp, greqs, &tt, &status ); 
  }
  PetscFree(greqs);       greqs = NULL;
  if(ngdata){ PetscFree(ngdata); ngdata = NULL; }
  
  // send real nodes
  ierr = PetscMalloc(nlp*sizeof(MPI_Request),&lreqs); CHKERRQ(ierr);
  for( xx = master, tt = 0 ; xx < endp ; xx += lfact, tt++ ) {
    MPI_Isend(nldata, loclszi, MPI_INT, xx/lfact, tag3, lcomm, &lreqs[tt]);
  }
  // new grid - receive real nodes
  const int nnodespart = ntlocpart + gh_fgid1__flid1_proc1.getCount();
  assert( mype != endp-lfact || ttt == nnodespart );
  if( nlbytes ){ ierr = PetscMalloc(nlbytes,&buff); CHKERRQ(ierr); }
  else buff = NULL;
  
  // create grid
  PromAggGrid *aggGrid = NULL; PromGeoGrid *geoGrid = NULL; 
  if( mgType_ == PROM_GEOMG ) {
    if( iscondndf ) {
      geoGrid = new PromGeoGrid( options_, perf_mon_,
				 dynamic_cast<PromGeoGrid*>(lastGrid), factor,
				 nli, buff, tag3, gh_fgid1__flid1_proc1,
				 ghostlist,ntlocpart, nnodespart,Comm_,ndof);
    }
    else {
      SETERRQ(1,"GEOMG w/ non-const ndf\n");
    }
    *newGrid = dynamic_cast<PromGrid*>(geoGrid);          CHKERRQ(ierr);      
  }
  else if( mgType_ == PROM_AGGMG ) {
    if( iscondndf ) {
      aggGrid = new PromAggGrid(options_, perf_mon_,
				dynamic_cast<PromAggGrid*>(lastGrid), factor,
				nli, buff, tag3, gh_fgid1__flid1_proc1,
				ghostlist, ntlocpart, nnodespart, Comm_,ndof);
    }
    else {
      aggGrid = new PromAggGrid( options_, perf_mon_,
				 dynamic_cast<PromAggGrid*>(lastGrid), factor,
				 nli, buff, tag3, gh_fgid1__flid1_proc1,
				 ghostlist, ntlocpart, nnodespart, nLocPartEq, 
				 Comm_ );
    }
    *newGrid = dynamic_cast<PromGrid*>(aggGrid);          CHKERRQ(ierr);
  }
  else assert(0);
  CHKNEW( *newGrid, sizeof(PromGrid) );    
  if(buff) { PetscFree(buff); buff = NULL; }

  // clean up 
  for( xx = master ; xx < endp ; xx += lfact ){
    MPI_Waitany( nlp, lreqs, &pid, &status ); 
  }
  PetscFree(lreqs);
  if(nldata != NULL) { ierr = PetscFree(nldata); CHKERRQ(ierr); }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus::FinalizeSymbolic"
/* Prometheus::FinalizeSymbolic ************************************************
 *
 *  - Finalize grids (MGFN)
 *
 *   INPUT:
 *
 *   SIDE EFFECTS:
 *     - sets data structues for numeric work
 *
 *   PRECOND:   '' & 'finalized' == FALSE & 
 *              !'top_grid' or 'top_grid' is symbolic and exact adjac.
 *   POSTCOND:  'finalized' == TRUE  & 'top_grid' is numeric and exact adjac. 
 *
 *  RETURN: 
 *    - PETSc error code
 */  
int Prometheus::FinalizeSymbolic()
{ 
  // finish (allocate data) when necc. 
  int ierr,np=Comm_.np(),mype=Comm_.mype(),verb=(options_.verbose_>1);
  MPI_Comm comm = Comm_.MPIComm();
  const int lev = top_grid_ ? top_grid_->getLevel() : 0;
  if( !grid0_ ) SETERRQ(1,"'MGFN' called with out init. - call 'MGIN'\n");
  if( finalized_ ) {  
    PetscPrintf(comm,"[%d]%s MGFN called after finalize\n",mype,__FUNCT__);
    return 0;
  }  
  ierr = perf_mon_.EventBegin( prom_logs_[LOG_PROM] ); CHKERRQ(ierr);
  // set up numeric data structures
  if(verb)PetscPrintf(comm,"[%d]%s Final MakeNumeric(%d)...\n", 
			  mype,__FUNCT__,lev);
  if( !top_grid_ ) { // one level solve
    if(verb)PetscPrintf(comm,"[%d]%s One level solv\n",mype,__FUNCT__);
    top_grid_ = grid0_;
  }
  if( top_grid_ != grid0_ && top_grid_->factor_ != np ) {
    if(verb)PetscPrintf(comm,"[%d]%s top grid not reduced\n",mype,__FUNCT__);
    bool b;
    ierr=options_.HasOptionName("-prometheus_use_parallel_top_solver",&b);
    CHKERRQ(ierr);
    if( b ) {
      top_grid_->stiffness_->ConvertToSuperLU(); CHKERRQ(ierr);
    }
    else {
      ierr = top_grid_->RepartitionTop( Comm_ ); CHKERRQ(ierr);
    }
  }
  // numeric
  ierr = top_grid_->PrepForSolve();    CHKERRQ(ierr); 
  // clean up symbolic buffers (ugly) 
  ierr = PromGeoNode::clearBuffers(); CHKERRQ(ierr); 
  ierr = perf_mon_.EventEnd( prom_logs_[LOG_PROM]); CHKERRQ(ierr);
  finalized_ = TRUE;
  PetscMemoryGetCurrentUsage(&rss);
  //if(verb)PetscPrintf(comm,"[%d]%s done %d grids (%d Mb)\n",mype,__FUNCT__, lev+1,(int)(rss/1.e6+.5));
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus::Archive"
/* Prometheus::Archive ***********************************************
 *
 *  
 *  RETURN:
 *    - PETSc error code 
 */
int Prometheus::Archive( FILE *file, const PromArchiveType type )
{
  int ierr,numlevels,ii;

  // do Prometheus -- more TODO!!!: mgType_
  if( type == PROM_PRINT ) {
    numlevels = getNumLevels();
    fprintf( file,"Prometheus: numlevels=%d\n",numlevels);
    
  }
  else if( type == PROM_WRITE ) {
    numlevels = getNumLevels();
    fprintf( file, "numlevels=%d\n", numlevels );
  }
  else {
    assert( type == PROM_READ );
    fscanf( file, "numlevels=%d\n", &numlevels);
    finalized_ = TRUE;
  }
  
  // do grids  
  PromGrid *grid = NULL;
  for( ii = 0, grid = grid0_ ; ii < numlevels  ; ii++ ){    
    if( type == PROM_WRITE ) { 
      assert(grid!=NULL); 
      fprintf(file,"nlocal=%d nNodes=%d nglobal=%d fact=%d\n",
	      grid->nlocal(),grid->nNodes_,grid->nglobal(),grid->factor_);
      ierr = grid->Archive( file, type );                        CHKERRQ(ierr);
      grid = dynamic_cast<PromGrid*>(grid->nextgrid_); 
      assert( grid || (ii+1) == numlevels );
    } 
    else if( type == PROM_READ ) { // READ
      int nn, nl, nt, fact, kk;
      kk = fscanf( file, "nLocalNd=%d nNodes=%d nGlobalNd=%d fact=%d\n",
		   &nl, &nn, &nt, &fact );
      if( kk != 4 ) SETERRQ(1,"fscanf != 4");
      /*tempgrid = new PromGrid(file,nl,nn,nt,fact,comm,&ierr);
	CHKNEW( tempgrid, sizeof(PromGrid) );                   CHKERRQ(ierr);
	if( !grid0_ ) { grid0_ = grid = tempgrid; } // first grid
	else{
	assert( tempgrid && grid && !grid->nextgrid_ && !tempgrid->prevgrid_ );
	grid->nextgrid_ = tempgrid;
	tempgrid->prevgrid_ = grid;
	grid = tempgrid;
	}*/
    }
  } 
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus::UpdateGeometry"
/* Prometheus::UpdateGeometry *******************************************
 *
 *  - update geometry data for P_0 in AMG methods. sets coord = crds+deltas
 *
 *   INPUT:
 *     - crds[3*grid->nNodes_]: original coords
 *     - deltas: current displacements (NULL?)
 *
 *   SIDE EFFECTS:
 *
 *  RETURN:
 *    - PETSc error code 
 */
int Prometheus::UpdateGeometry( double * const crds, double * const deltas )
{
  int ierr,mype=Comm_.mype(),verb=(options_.verbose_>1);
  MPI_Comm comm = Comm_.MPIComm();

  if( grid0_ == NULL ) {
    PetscPrintf(comm,"[%d]%s no fine grid\n",mype,__FUNCT__);
  }
  else if( mgType_ == PROM_AGGMG ) {
    const PromAggGrid *grid = dynamic_cast<PromAggGrid*>(grid0_);
    const PromAggGrid *nnext = dynamic_cast<PromAggGrid*>(grid->nextgrid_);
    if( grid != NULL ) {
      ierr = perf_mon_.EventBegin(prom_logs_[LOG_SLES_SETUP]); CHKERRQ(ierr);
      if(verb)PetscPrintf(comm,"[%d]%s SetCoords\n",mype,__FUNCT__);
      // initialize SetCoords w/ new deltas
      ierr = grid->SetCoords( crds, 3, deltas );  CHKERRQ(ierr);
      while( nnext != NULL ) {
	if( grid->isActive() ) {
	  if(verb)PetscPrintf( comm,"[%d]%s QRShapes(%d)\n",
			       mype,__FUNCT__,grid->getLevel());
	  ierr = nnext->QRShapes( grid ); CHKERRQ(ierr);
	}
	// clear marks and Bs
	for( int xx = 0 ; xx < grid->nNodes_  ; xx++ ){
	  PromAggNode *fine = dynamic_cast<PromAggNode*>(grid->nodes_[xx]);
	  fine->mark( FALSE ); 	fine->SetBsz(0); // no need any more
	}
	// increment
	grid = nnext;
	nnext = dynamic_cast<PromAggGrid*>(grid->nextgrid_);
      }
      ierr = perf_mon_.EventEnd( prom_logs_[LOG_SLES_SETUP] ); CHKERRQ(ierr);
    }
  }
  else { // geo mg
    const PromGeoGrid *grid = dynamic_cast<PromGeoGrid*>(grid0_);
    assert(grid != NULL);
    ierr = grid->SetCoords( crds, 3, deltas );  CHKERRQ(ierr);
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "Prometheus::nBlocksWithGlobalDOF"
/* int Prometheus::nBlocksWithGlobalDOF ***************************************
 *
 *  INPUT:
 *    - gl: number of global DOFs
 *
 *  RETURN:
 *    - recommended number of global blocks 
 *    - all procs return same if grid0_->nBlocks_ are same
 */
int Prometheus::nBlocksWithGlobalDOF( const int nneq ) const
{
  assert(this!=NULL);  assert(grid0_!=NULL);
  int nb0, nbtot, nneq0; bool flag;
  
  options_.GetOptionInt( "-prometheus_blocks", &nb0, &flag );
  if( !flag ) {
    options_.GetOptionInt( "-prometheus_nodes_per_block",
			    &nb0, &flag );
    if( !flag ) return 0;
    nb0 = grid0_->nglobal() / nb0; // global # blocks
    if( nb0 <= 0 ) nb0 = 1;
  }
  nneq0 = grid0_->nglobalEq();    assert(nneq<=nneq0);
  nbtot = (int)( (float)nb0 * ((float)nneq/(float)nneq0) + 0.5 );
  if( nbtot < 1 ) nbtot = 1;

  return nbtot;
}

#undef __FUNCT__ 
#define __FUNCT__ "Prometheus::SetMatrixWithNewData"
/* int Prometheus::SetMatrixWithNewData ***************************************
 *
 *  - For fine grid only.  Calles assemble done and scales matrix
 *
 *  INPUT:
 *  RETURN:
 *    - recommended number of global blocks
 */
int Prometheus::SetMatrixWithNewData( PromMatrix *AA )
{
  PromGrid *grid0 = grid0_;    
  int ierr,ii,xx,jj,geq,mype=Comm_.mype(); 

  PetscFunctionBegin;
  ierr = AA->Assembly(); CHKERRQ(ierr);
  //
  if( left_ == NULL && right_ == NULL ) {
    bool flag;
    ierr = options_.HasOptionName( "-prometheus_scale_matrix", &flag );
    CHKERRQ(ierr); // complex?
    if( flag ){
      left_ = new PromVector( AA->map_, AA->isComplex() ? 2 : 1 );
      right_ = left_;
      //
      double t1, t2, *diag;
      PromVector *work = new PromVector( AA->map_, AA->isComplex() ? 2 : 1 );
      ierr = AA->PromMatrix::GetDiagonal( work );      CHKERRQ(ierr);
      ierr = work->GetArray( &diag );                  CHKERRQ(ierr);
      for( xx = ii = 0, geq = grid0->my0Eq() ; 
	   xx < grid0->nLocalNd_ ; xx++ ) {
	int ndf = grid0->nodes_[xx]->getNDF();
	for( jj = 0 ; jj < ndf ; jj++, ii++, geq++ ){
	  assert(diag[ii] > 0.);
	  t1 = sqrt(diag[ii]); t2 = 1.0 / t1;
	  ierr = left_->SetValues( 1, &geq, &t2, INSERT_VALUES ); 
	  CHKERRQ(ierr);
	}
      }
      assert(geq==grid0->proc_geq()[mype+1]);
      ierr = work->RestoreArray( &diag );        CHKERRQ(ierr);
      delete work;
    }
    else {
      ierr = options_.HasOptionName( "-prometheus_left_scale_matrix", &flag );
      CHKERRQ(ierr); // complex?
      if( flag ){
	left_ = new PromVector( AA->map_, AA->isComplex() ? 2 : 1 );
	//
	double t1, t2, *diag;
	PromVector *work = new PromVector( AA->map_, AA->isComplex() ? 2 : 1 );
	ierr = AA->PromMatrix::GetDiagonal( work );      CHKERRQ(ierr);
	ierr = work->GetArray( &diag );                  CHKERRQ(ierr);
	for( xx = ii = 0, geq = grid0->my0Eq() ; 
	     xx < grid0->nLocalNd_ ; xx++ ) {
	  int ndf = grid0->nodes_[xx]->getNDF();
	  for( jj = 0 ; jj < ndf ; jj++, ii++, geq++ ){
	    t1 = diag[ii]; t2 = 1.0 / t1;
	    ierr = left_->SetValues( 1, &geq, &t2, INSERT_VALUES ); 
	    CHKERRQ(ierr);
	  }
	}
	assert(geq==grid0->proc_geq()[mype+1]);
	ierr = work->RestoreArray( &diag );        CHKERRQ(ierr);
	delete work;
      }
      else {
	ierr = options_.HasOptionName( "-prometheus_right_scale_matrix", &flag );
	CHKERRQ(ierr); // complex?
	if( flag ){
	  right_ = new PromVector( AA->map_, AA->isComplex() ? 2 : 1 );
	  //
	  double t1, t2, *diag;
	  PromVector *work = new PromVector( AA->map_, AA->isComplex() ? 2 : 1 );
	  ierr = AA->PromMatrix::GetDiagonal( work );      CHKERRQ(ierr);
	  ierr = work->GetArray( &diag );                  CHKERRQ(ierr);
	  for( xx = ii = 0, geq = grid0->my0Eq() ; 
	       xx < grid0->nLocalNd_ ; xx++ ) {
	    int ndf = grid0->nodes_[xx]->getNDF();
	    for( jj = 0 ; jj < ndf ; jj++, ii++, geq++ ){
	      t1 = diag[ii]; t2 = 1.0 / t1;
	      ierr = right_->SetValues( 1, &geq, &t2, INSERT_VALUES ); 
	      CHKERRQ(ierr);
	    }
	  }
	  assert(geq==grid0->proc_geq()[mype+1]);
	  ierr = work->RestoreArray( &diag );        CHKERRQ(ierr);
	  delete work;
	}
      }
    }
  }
  // make scaling vecs and scale matrix
  if( left_ != NULL || right_ != NULL ) { 
    ierr = AA->Scale( left_, right_ );           CHKERRQ(ierr);
  }
  //
  ierr = grid0_->FineGridInitWithA();    CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus::CreateFineGrid"
/* Prometheus::CreateFineGrid ************************************************
 *
 *  Prometheus constructor
 *
 *  INPUT:
 *     -npLoc: number of local nodes here
 *     -lid_ndof:
 *     -ndf:
 *     -gid_ghostLid:
 *     -elemList:
 *
 *  SIDE EFFECTS:
 *    - 
 *
 *  RETURN:
 */
int Prometheus::CreateFineGrid(const int npLoc, const int lid_ndof[],
			       const int ndf, const PromTable<int> &gid_ghostLid,
			       const PromList<int*> *elemList )
{
  int ierr,node_per_block;
  // factory for fine grid
  if( mgType_ == PROM_AGGMG ){
    if( ndf > 0 ) {
      grid0_ = new PromAggGrid( options_, perf_mon_, &parBank_, npLoc, 
				gid_ghostLid, elemList, Comm_, ndf );
    }
    else {
      grid0_ = new PromAggGrid( options_, perf_mon_, &parBank_, npLoc, 
				gid_ghostLid, elemList, Comm_, lid_ndof );
    }
  }
  else{
    // 
    grid0_ = new PromGeoGrid( options_, perf_mon_, &parBank_, npLoc, 
			      gid_ghostLid, elemList, Comm_, ndf );
  }

  // collect idex data for nodes
  ierr = grid0_->SetIndex( gid_ghostLid ); CHKERRQ(ierr);
  
  // make boundry node deps, get ghost data
  ierr = grid0_->GetGhostData();                      CHKERRQ(ierr);
  top_grid_ = grid0_; // cache top_grid for new grids
  
  // blocks
  int nblcks, ngnd = grid0_->nglobal(); assert(npLoc == grid0_->nlocal());
  bool flag;        assert(grid0_->nBlocks_ == 0);
  ierr = options_.GetOptionInt("-prometheus_nodes_per_block", &node_per_block, &flag ); 
  CHKERRQ(ierr);
  if( flag ) {
    if( node_per_block > 0 ) nblcks = npLoc / node_per_block + 1;
    else nblcks = 1;
  }
  else { 
    options_.GetOptionInt( "-prometheus_blocks", &nblcks, &flag );
    if( flag ) {
      if(nblcks>0) {
	node_per_block = ngnd/nblcks + 1;
	nblcks = npLoc / node_per_block + 1;
      }
      else nblcks = 1;
    }
  }
  if( flag ) grid0_->nBlocks_ = nblcks;
  //
  return 0;
}  

#undef __FUNCT__
#define __FUNCT__ "PromContext::TInvert" 
/* PromContext::TInvert ************************************************
 *
 *  - explicit inverse of transpose
 *
 *  INPUT: 
 *    - src: (in)
 *    - dest: (out)
 *    - NN: (in)
 *    - LDAt: (in)
 *
 *  RETURN:
 *   - PETSc error code
 */
int PromContext::TInvert( const double src[], double dest[], const int NN, 
			  const int LDAt /*=-1*/) 
{
  int       ii,jj,kk,k,errors=0,errors2=0,buf[2*PROM_MAXNDF];  const int ndm=NN; 
  double    dd,tol=1.e-200;
  const int LDA = (LDAt < 1) ? NN : LDAt;

  // copy src into dest
  for( kk = ii = 0 ; ii < NN ; ii++ ) {
    for( jj = 0, k = ii ; jj < NN ; jj++, kk++, k += LDA ) {
      dest[kk] = src[k]; // transpose
    }
  }
  // zero out 
  for( kk = 0 ; kk < NN ; kk++ ) {
    const int kn = kk*ndm;
    if( fabs(dest[kn+kk]) < tol ) {
      //PromContext::printArr(stderr,src,NN*NN,"ERROR zero pivot [%d]",NN);
      buf[errors++] = kk; //errors2++;
      for(jj=0 ; jj<NN ; jj++){
	if(jj!=kk) dest[kn+jj] = dest[jj*ndm+kk] = 0.;
	else dest[kn+jj] = 2.;
      }
    }
  }
  buf[errors] = -1;

  // in place inverse
  for( kk = 0, errors = 0 ; kk < NN ; kk++ ) {
    const int kn = kk*ndm, knk = kn + kk;
    if( (dd=dest[knk]) == 0.0 || buf[errors] == kk ) {
      //dest[knk] = 0.; // / dd;//*PROM_DIAG_FACTOR;
      errors2++; errors++; 
    }
    else {
      dd = 1./dd;
      for( jj = 0, k = kn ; jj < NN ; jj++, k++ ) dest[k] = -dest[k]*dd;
      for( ii = 0 ; ii < NN ; ii++ ) {
	const int in = ii*ndm, ink = in + kk;
	if(kk != ii) {
	  for( jj = 0, k = kn ; jj < NN ; jj++, k++ ) {
	    if( kk != jj ) {
	      dest[in + jj] += dest[ink] * dest[k];
	    }
	  }
	}
	dest[ink] = dest[ink]*dd;
      }
      dest[knk] = dd;
    }
  }

  return errors2;
}

#undef __FUNCT__  
#define __FUNCT__ "Prometheus::getNumLevels"
/* Prometheus::getNumLevels *************************************************
 *
 *  - returns number of levels currently allocated
 */
int Prometheus::getNumLevels ()const
{
  assert(this!=NULL);
  PromGrid* g = grid0_; 
  //assert(g);
  int i = 0;
  while( g != NULL ){ g = g->nextgrid_; i++;  }  
  return i;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::GetLocalNodeRow" 
/* PromMatrix::GetLocalNodeRow *********************************************
 *
 *  - get rows (of node) of Matrix.
 *
 *  INPUT:
 *    - MyRow: block row (local index)
 *    - len: 
 *    - ncols: size of row. (out)
 *    - adjacs[len]: column global nodes in row.(out) -- can be NULL
 *    - Ai[len]: buffer to hold row poiters, [B,A] (out) -- can be NULL
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromMatrix::GetLocalNodeRow(const int MyRow, const int len, int &ncols, 
				int adjacs[]/*=NULL*/, double **Ai/*=NULL*/)
  const
{
  int ierr;
#if defined(PROM_USE_PETSC)  
  if( map_.isConstNDF() ) {
    if( isSymmetricMatrix() ) { 
      SETERRQ(-1,"not supported"); // getrow not used on symm.
      //ierr = GetSymmetricLocal_constBlk_NodeRow(MyRow, len, ncols, adjacs, Ai);
      //CHKERRQ(ierr);
    }
    else {
      if( isComplex() ){
	ierr = GetNodeScalarRow_PETSc(MyRow, len, ncols, adjacs, Ai);
	CHKERRQ(ierr);
      }
      else {
	ierr=GetLocal_constBlk_NodeRow(MyRow,len,ncols,adjacs,Ai);CHKERRQ(ierr);
      }
    }
  }
  else { // VBR
    ierr = GetLocalVBRNodeRow( MyRow, len, ncols, adjacs, Ai );CHKERRQ(ierr);
  }
#elif defined(PROM_USE_PETRA)
  Epetra_VbrMatrix *mat = dynamic_cast<Epetra_VbrMatrix*>(pmat_);
  if( mat != NULL ) {
    int fndf, *LDAs, *cndf; // Ai may be NULL
    ierr = mat->ExtractMyBlockRowPointers( MyRow, len, fndf, ncols, adjacs, 
                                           cndf, LDAs, Ai );
    CHKERRQ(ierr);
  }
  else {
    Epetra_CrsMatrix *mat = dynamic_cast<Epetra_CrsMatrix*>(pmat_);
    assert(mat!=NULL); 
    double *Values; int *Indices;
    ierr = mat->ExtractMyRowView( MyRow, ncols, Values, Indices ); 
    CHKERRQ(ierr);
    for(int ii=0;ii<ncols;ii++){
      adjacs[ii] = Indices[ii]; 
      Ai[ii] = &Values[ii];
    }
  }
#endif

  return 0;  
} 

#if defined(PROM_USE_PETSC)

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::PromMatrix" 
/* PromMatrix::PromMatrix *****************************************
 *
 *  - constructor (can be non-const. ndf, square)
 */
PromMatrix::PromMatrix( const PromGrid *grid, int nA,int *pnA,int nB,
			int *pnB, int fact /*=1*/ ) : 
  PromMatrix_base(*grid), mat_(NULL), full_mat_(NULL), array_(NULL), 
  varray_(NULL), ghostGID_BID_(NULL), ghostGID_ndf_(NULL), 
  ghostBID_GID_(NULL), varr_sz_(0), malled_(1)
{
  assert(this!=NULL);   
  PetscFunctionBegin;
  MPI_Comm wcomm = map_.comm_.MPIComm(); 
  if( !map_.isConstNDF() ) {
    // VBR matrix
    int ii,jj,kk,dof,nloc=map_.nlocal();
    const int *lid_eq = map_.id_leq_, mype = map_.comm_.mype();
    // get lowest common denominator (easy way)
    for( ii = 0, kk = 10000 ; ii < nloc ; ii++ ) {
      dof = lid_eq[ii+1] - lid_eq[ii];
      if( dof < kk ) kk = dof; 
    }
    MPI_Allreduce( &kk, &dof, 1, MPI_INT, MPI_MIN, wcomm );
    const int bs = dof*fact,lbs=dof,maxf=PROM_MAXNDF/dof; // traditional 'bs'
    const int nMatNodes = map_.nlocalEq()/lbs; assert(map_.nlocalEq()%lbs==0);
    // check for true lowest common denom.
    for( ii = 0 ; ii < nloc ; ii++ ) {
      dof = lid_eq[ii+1] - lid_eq[ii];
      if( dof%lbs != 0 ){ 
	PetscPrintf(PETSC_COMM_SELF,"[%d]%s ERROR dof=%d lbs(%d)=%d nnz=%d\n",
		    mype,__FUNCT__,dof,ii,lbs,pnA[ii]);
	/* since function returns void cannot use PetscFunctionReturn(); */
	PetscStackPop; 
	return;
      }
    }
    // set ndf ... tables
    PromTable<int> *g_bid, *g_ndf, *b_gid; int bsz = 4*(nloc-nloc)/3+11;
    g_bid = ghostGID_BID_ = new PromTable<int>(bsz);
    g_ndf = ghostGID_ndf_ = new PromTable<int>(bsz);
    b_gid = ghostBID_GID_ = new PromTable<int>(bsz);
    // fill helpers 'ghostGID_BID_' 'ghostGID_ndf_' 'ghostBID_GID_' 
    TablePos tpos = grid->ghosts_.getHeadPosition(); 
    while( tpos ){
      PromNode *n1 = grid->ghosts_.getNext( tpos ); 
      int geq=n1->geq(),bid=geq/lbs,gid=n1->index_.global(),ndf=n1->getNDF(); 
      g_bid->add( bid+1, gid+1 );         // 'ghostGID_BID_'
      g_ndf->add( ndf, gid+1 );           // 'ghostGID_ndf_'
      jj = ndf/lbs; assert(jj>0 && jj<8);
      for(ii=0;ii<jj;ii++){
	kk = b_gid->add(gid+1, bid+1+ii); assert(kk==1);  // 'ghostBID_GID_'
      }
    }
    // get nnz/onnz
    int *nnz=NULL,*onnz=NULL,maxNNZ=grid->nlocalEq()/lbs,maxO=grid->nglobalEq()/lbs-maxNNZ;
    if( nMatNodes > 0 ) {
      assert( pnA != NULL ); PetscMalloc( sizeof(int)*(nMatNodes), &nnz ); 
      assert( pnB != NULL ); PetscMalloc( sizeof(int)*(nMatNodes), &onnz );
      for( ii = jj = kk = 0 ; ii < nloc ; ii++ ) {
	dof = (lid_eq[ii+1] - lid_eq[ii]);
	int olsz=pnB[ii],lsz=pnA[ii],nlb=dof/lbs;
	for( int i = 0 ; i < nlb ; i++, jj++, kk++ ) { 
	  nnz[jj] = lsz*maxf; onnz[kk] = olsz*maxf; 
	  if( nnz[jj] > maxNNZ ) nnz[jj] = maxNNZ;
	  if( onnz[jj] > maxO ) onnz[jj] = maxO;
	}
      }
      assert(jj==nMatNodes);assert(jj==kk);
    }
    if( bs > 1 ) {
      MatCreateMPIBAIJ(wcomm, bs, map_.nlocalEq()*fact, 
		       map_.nlocalEq()*fact, map_.nglobalEq()*fact,
		       map_.nglobalEq()*fact, fact*nA, nnz, fact*nB, onnz, &mat_);
    }
    else {
      MatCreateMPIAIJ(wcomm, map_.nlocalEq()*fact, map_.nlocalEq()*fact, 
		      map_.nglobalEq()*fact, map_.nglobalEq()*fact, nA, 
		      nnz, nB, onnz, &mat_);
    }
    if( nnz != NULL ) PetscFree(nnz);
    if( onnz != NULL ) PetscFree(onnz);
  }
  else {
    SetUp( nA, pnA, nB, pnB, map_.nlocalEq(), map_.nglobalEq(), fact );
  }
  
  PetscStackPop; /* since function returns void cannot use PetscFunctionReturn(); */
}

/* PromMatrix::PromMatrix *****************************************
 *
 *  - constructor (const. ndf, can be rectangular)
 */
PromMatrix::PromMatrix( const PromMap &map, int nA,int *pnA,int nB,int *pnB, 
			int lcols /*= -1*/, int gcols /*= -1*/, int fact /*=1*/): 
  PromMatrix_base( map ), mat_(NULL), full_mat_(NULL), array_(NULL),varray_(NULL),
  ghostGID_BID_(NULL),ghostGID_ndf_(NULL),ghostBID_GID_(NULL),varr_sz_(0),
  malled_(1)
{
  assert(this!=NULL); 
  if( lcols == -1 ){ lcols = map_.nlocalEq(); gcols = map_.nglobalEq(); }
  SetUp( nA, pnA, nB, pnB, lcols, gcols, fact );
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::SetUp" 
/* PromMatrix::SetUp *****************************************
 *
 *  - constructor (const. ndf, can be rectangular)
 */
int PromMatrix::SetUp( int nA,int *pnA,int nB,int *pnB, int lcols/*=-1*/,
		       int gcols/*=-1*/, int fact /*=1*/ ) 
{
  int ierr;
  assert( map_.comm_.MPIComm() != MPI_COMM_NULL );
  MPI_Comm wcomm = map_.comm_.MPIComm(); assert(fact==1 || fact==2);

  assert( map_.isConstNDF() );
  if( fact*map_.bsize_ == 1 ) {
    //! scalar matrix -- need to provide left and right comm
    ierr = MatCreateMPIAIJ( wcomm, fact*map_.nlocalEq(), fact*lcols, 
			    fact*map_.nglobalEq(), fact*gcols, 
			    nA, pnA, nB, pnB, &mat_);
    if(ierr != 0) mat_ = NULL;  
  }
  else {
    //! simple for const block matrices
    ierr = MatCreateMPIBAIJ(wcomm, fact*map_.bsize_, fact*map_.nlocalEq(),
			    fact*lcols, fact*map_.nglobalEq(), fact*gcols,
			    nA, pnA, nB, pnB, &mat_ );
    if(ierr != 0) mat_ = NULL;
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::SetValues" 
/* PromMatrix::SetValues *****************************************
 *
 *  - 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromMatrix::SetValues( const int nrow, int *pgeqi, const int ncol, int *pcol,
			   double *vals, const int add_t )
{
  assert(this!=NULL); int ierr; 
  dirty_data_ = TRUE; // not collective!
  ierr = MatSetValues( mat_, nrow, pgeqi, ncol, pcol, vals, (InsertMode)add_t);
  CHKERRQ(ierr);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::SetValuesBlocked" 
/* PromMatrix::SetValuesBlocked *****************************************
 *
 *  - 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromMatrix::SetValuesBlocked(const int nrow, int *prows, const int ncol,
				 int *pcols, double *vals, const int add_t)
{
  assert(this!=NULL); 
  int ierr,bs;
  ierr = MatGetBlockSize( mat_, &bs ); CHKERRQ(ierr);

  dirty_data_ = TRUE; // not collective!
  if( map_.isConstNDF() ) {
    if( bs == 1 ) {
      ierr=MatSetValues(mat_,nrow,prows,ncol,pcols,vals,(InsertMode)add_t); 
      CHKERRQ( ierr );
    }
    else {
      ierr = MatSetValuesBlocked( mat_, nrow, prows, ncol, pcols, vals, 
				  (InsertMode)add_t );
      CHKERRQ( ierr );
    }
  }
  else{ 
    assert(!isComplex()); // avoid!
    // VBR: assumes rows are local
    assert(nrow<1024 && ncol<1024);
    const int my0=map_.my0Nd(),nloc=map_.nlocal(),next0=my0+nloc;
    const int *const lid_leq = map_.id_leq_, my0B=map_.my0Eq()/bs; 
    int nc,nr,ii,jj,gbid,gid,nblk=-1,lid,bgidrows[1024], bgidcols[1024]; 
    // rows
    for(jj=ii=0;ii<nrow;ii++){
      gid = prows[ii]; 
      if( gid < my0 || gid >= next0 ) {
	gbid = ghostGID_BID_->find(gid+1) - 1; assert(gbid>=0);
	int dof = ghostGID_ndf_->find(gid+1); assert(dof>0);
	nblk = dof/bs; // number of blocks
      }
      else {
	lid = gid - my0; nr = lid_leq[lid]; gbid = nr/bs + my0B; 
	nblk = (lid_leq[lid+1]/bs + my0B) - gbid ; // number of blocks
      }
      // expand id arr.
      for(int k = 0 ; k < nblk ; k++, jj++) bgidrows[jj] = gbid + k; 
    }
    nr = jj;
    // cols
    for(jj=ii=0;ii<ncol;ii++){
      gid = pcols[ii]; 
      if( gid < my0 || gid >= next0 ) {
	gbid = ghostGID_BID_->find(gid+1) - 1; assert(gbid>=0);
	int dof = ghostGID_ndf_->find(gid+1); assert(dof>0);
	nblk = dof/bs; // number of blocks
      }
      else{
	lid = gid - my0; nc = lid_leq[lid]; gbid = nc/bs + my0B; assert(nc%bs==0); 
	nblk = (lid_leq[lid+1]/bs + my0B) - gbid;
      }
      for(int k = 0 ; k < nblk ; k++, jj++ ) bgidcols[jj] = gbid + k;
    }
    nc = jj;
    ierr = MatSetValuesBlocked( mat_, nr, bgidrows, nc, bgidcols, vals,
				(InsertMode)add_t);
    CHKERRQ( ierr );
  }
  return 0;
}
#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::getLidWithLBid" 
/* PromMatrix::getLidWithLBid *************************************
 *
*  - find lid with (internal) blid
 *
 *  INPUT:
 *    - blid: matrix block number
 *    - bs: matrix block size
 *    - nloc: num local nodes - size of 'lid_leq'
 *    - lid_leq: [nloc] map from local node ID to local equation
 *    - fact: 1 or 2
 *    - rem: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - lid
 */
int PromMatrix::getLidWithLBid(const int blid, const int bs, const int nloc, 
			       const int lid_leq[], const int fact, int &rem )
{
  const int leq = blid*bs/fact; int e1,e0,tt,ff=1,lid = blid; assert(bs%fact==0);
  if( lid >= nloc ) lid = nloc - 1;
  while( leq < lid_leq[lid] || leq >= lid_leq[lid+1] ) {
    e1 = lid_leq[lid+1]; e0 = lid_leq[lid];
    if( leq < (e1-10) ){
      tt = (e1-leq)/(e1-e0)/ff;
      if( tt < 1 ) lid--; 
      else{ 
	if( tt > lid ) lid = lid/2;
	else { lid -= tt; ff++; }
      }
    }
    else if( leq < e1 ) lid--;
    else if( leq > (e1+10) ){
      tt = (leq-e1)/(e1-e0)/ff;
      if( tt < 1 ) lid++;
      else {
	if( (lid+tt) >= nloc ) lid = (lid+nloc)/2;
	else { lid += tt; ff++; }
      }
    }
    else lid++;
#if defined(PETSC_USE_BOPT_g)
    if(lid<0||lid==nloc)SETERRQ3(1,"Failed lookup. bgid=%d lid=%d leq=%d",
				 blid,lid,leq);
#endif
  }
  rem = fact*(leq - lid_leq[lid])/bs;
  return lid;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::GetLocalVBRNodeRow" 
/* PromMatrix::GetLocalVBRNodeRow *************************************
 *
 *  - get rows (of node) of Matrix.
 *
 *  INPUT:
 *    - rlid: block row (local index)
 *    - len: 
 *    - ncols: size of row. (out)
 *    - adjacs[len]: column global nodes in row.(out) -- can be NULL
 *    - Ai[len]: buffer to hold row poiters, [B,A] (out) -- can be NULL
 *
 *  SIDE EFFECTS:
 *    - uses 'varr_sz_' and  'varray_'
 *
 *  RETURN:
 *      - PETSc error code
 */
int PromMatrix::GetLocalVBRNodeRow( const int rlid, const int len, 
				    int &ncols, int adjacs[], double **Ai)
  const
{
  int        ierr,jjj,jb,ii,nn,ndf,gid,rem;
  const int  my0=map_.my0Nd(),nloc=map_.nlocal(), fact = isComplex() ? 2 : 1;  
  ierr = MatGetBlockSize(mat_,&ii); CHKERRQ(ierr); assert(rlid>=0&&rlid<nloc);
  const int  bs=ii,my0B=map_.my0Eq()/(bs/fact),next0B=my0B+map_.nlocalEq()/(bs/fact);
  const int *const lid_leq=map_.id_leq_; assert(lid_leq!=NULL);
  const int  leqi=lid_leq[rlid],ndfi=fact*(lid_leq[rlid+1]-leqi); // stride of out
  const int  rblid=fact*leqi/bs,nblki=ndfi/bs;
  double    *Ais[PROM_MAXNDF][PROM_MAXNIEGH]; assert(bs%fact==0);
  int        adjacss[PROM_MAXNDF][PROM_MAXNIEGH],nbc[PROM_MAXNDF],buffer[2*PROM_MAXNIEGH];
  PromTable<int> gid_ndf_or_jind(73);

  // get raw data
  assert(nblki > 0);
  for( ii = jb = nn = 0 ; ii < nblki ; ii++ ){
    if( isSymmetricMatrix() ) {
      SETERRQ(-1,"not supported"); // getrow not used on symm.
      //ierr = GetSymmetricLocal_constBlk_NodeRow(rblid+ii, PROM_MAXNIEGH, nbc[ii], adjacss[ii], (Ai==NULL) ? NULL :  Ais[ii]); CHKERRQ(ierr);
    }
    else{
      ierr = GetLocal_constBlk_NodeRow( rblid+ii, PROM_MAXNIEGH, nbc[ii],
					adjacss[ii], (Ai==NULL) ? NULL: Ais[ii]);
      CHKERRQ(ierr);
    }

    for( jjj = 0 ; jjj < nbc[ii] ; jjj++ ){
      const int bcid = adjacss[ii][jjj];
      if( bcid < my0B || bcid >= next0B ) { // ghost
	gid = ghostBID_GID_->find(bcid+1) - 1;
	if( gid < 0 ) SETERRQ1(1,"error bcid %d not found",bcid);
	ndf = fact*ghostGID_ndf_->find(gid+1); assert(ndf>0);
      }
      else {
	int lid = getLidWithLBid(bcid-my0B,bs,nloc,lid_leq,fact,rem);
	gid = lid + my0;
	ndf = fact*(lid_leq[lid+1]-lid_leq[lid]); 
	assert(rem < ndf/bs);
      }
      if( !gid_ndf_or_jind.find( gid + 1 ) ) {
	gid_ndf_or_jind.add( ndf, gid + 1 );
	buffer[2*nn] = gid; buffer[2*nn+1] = ndf;
	if(adjacs!=NULL && Ai==NULL) adjacs[nn] = gid; // out (w/o Aii)
	nn++; jb += ndf; 
      }
    }
  }
  ncols = nn; // (out)
  if( Ai == NULL ) return 0;       assert( adjacs != NULL );
  const int nnz=ndfi*jb; assert(jb>0);  assert(nnz>0); 
  qsort( buffer, ncols, 2*sizeof(int), prom_intcompar );
  // buffers
  if( varr_sz_ < nnz+1 ) {
    varr_sz_ = nnz+1;
    if(varray_ != NULL) PetscFree( varray_ );
    ierr = PetscMalloc( (nnz+1)*sizeof(double), &varray_ );    CHKERRQ(ierr);
  }
  PetscMemzero( varray_, nnz*sizeof(double) ); varray_[nnz] = PROM_BIG_NUM;
  // collect gids
  for( nn = jb = 0 ; nn < ncols ; nn++ ){
    gid = buffer[2*nn]; ndf = buffer[2*nn+1]; 
    Ai[nn] = varray_ + ndfi*jb;      // out
    adjacs[nn] = gid; // out
    gid_ndf_or_jind.add( jb + 1, gid + 1 );
    jb += ndf;
  }
  assert(nnz == ndfi*jb);

  // copy in
  for( ii = 0 ; ii < nblki ; ii++ ){
    for( jjj = 0 ; jjj < nbc[ii] ; jjj++ ){
      const int bcid = adjacss[ii][jjj];
      if( bcid < my0B || bcid >= next0B ) { // ghost
	gid = ghostBID_GID_->find(bcid+1) - 1;	assert(gid>=0);
	int id = ghostGID_BID_->find(gid+1) - 1; assert(id>=0);
	rem = (bcid - id);
      }
      else {
	int lid = getLidWithLBid(bcid-my0B,bs,nloc,lid_leq,fact,rem);
	gid = lid + my0; 
	assert(rem < (fact*(lid_leq[lid+1] - lid_leq[lid]))/bs);
      }
      jb = gid_ndf_or_jind.find( gid + 1 ) - 1; assert(jb>=0);
      //
      nn = ndfi*(jb + rem*bs) + ii*bs; 
      double *const dest = varray_ + nn;
      const double *src = Ais[ii][jjj]; assert(src!=NULL);
      for( int j = 0 ; j < bs ; j++ ) {   // cell cols
	for( int i = 0 ; i < bs ; i++ ) {
	  int id = i + j*ndfi;
	  if( dest[id] != 0.0 ) SETERRQ2(1,"dest[%d] != 0.0 == %e",id,dest[id]);
	  dest[id] = *src++; // source is transposed
	} // i
      } // j 
    } // jjj
  } // ii
  assert(varray_[nnz] == PROM_BIG_NUM);

  return 0;
} 


#undef __FUNCT__  
#define __FUNCT__ "PromComplexMatrix::SetValuesBl_private"
/* PromComplexMatrix::SetValuesBl_private ****************************
 *
 * - assembly
 */ 
int PromComplexMatrix::SetValuesBl_private( const int nrow, int *prow, 
					    const int ncol, int *pcol,
					    double *vals, const int add_t,
					    const bool iscomplex )
{
  int ii,ierr,ci,cj,kk,jj,ndfi,ndfj,lid,gid,pt1,ndfj_tot,ndfi_tot,bs=0,geq;
  const int my0 = map_.my0Nd(), nloc = map_.nlocal();
  const bool isconst = map_.isConstNDF(), my0Eq = map_.my0Eq();
  const int *const lid_leq=map_.id_leq_; 
  double tmp,*pv,tvals[4096];           assert(add_t==ADD_VALUES);
  int pgidi[64],pgidj[64],ndfjs[PROM_MAXNIEGH]; 

  // get ndfjs
  if( isconst ) {
    ierr = MatGetBlockSize( mat_, &bs ); CHKERRQ(ierr); 
  }
  else assert(lid_leq!=NULL);
  for( ndfj_tot = jj = 0 ; jj < ncol ; jj++ ){
    gid = pcol[jj]; lid = gid - my0;
    if(isconst){ ndfj = bs; geq = gid*bs; }
    else if(lid >= 0 && lid < nloc){
      ndfj = 2*map_.locID_ndf(lid); 
      assert(ndfj==2*(lid_leq[lid+1]-lid_leq[lid]));
      geq = 2*(my0Eq + lid_leq[lid]);
    }
    else{ 
      ndfj = 2*ghostGID_ndf_->find(gid+1); 
      geq = bs*(ghostGID_BID_->find(gid+1)-1); assert(geq>=0);
    }
    ndfjs[jj] = ndfj; assert(ndfj>0);
    for( kk = 0 ; kk < ndfj ; kk++ ) pgidj[ndfj_tot++] = geq++;
  }
  // add in
  for( ndfi_tot = ci = ii = 0, pv = vals ; ii < nrow ; ii++ ){
    gid = prow[ii]; lid = gid - my0;
    if( isconst ) { ndfi = bs; geq = gid*bs; }
    else if( lid >= 0 && lid < nloc ){
      ndfi = 2*map_.locID_ndf(lid); geq = 2*(my0Eq + lid_leq[lid]);
    }
    else{
      ndfi = 2*ghostGID_ndf_->find(gid+1);
      geq = bs*(ghostGID_BID_->find(gid+1)-1); assert(geq>=0);
    }
    for( kk = 0 ; kk < ndfi ; kk++ ) pgidi[ndfi_tot++] = geq++;
    //
    ndfi = ndfi/2;
    for( int i = 0 ; i < ndfi ; i++, ci += 2*ndfj_tot ) {
      for( cj = jj = 0 ; jj < ncol ; jj++ ){
	ndfj = ndfjs[jj]/2;
	for( int j = 0 ; j < ndfj ; j++, cj += 2, pv++ ){
	  tmp = *pv; pt1 = ci + cj;  int pt2 = pt1 + ndfj_tot;
	  if(pt2>=4096)SETERRQ2(-1,"buffer too small %d > %d",pt2,4096);
	  if( iscomplex ) {
	    tvals[pt1] = tvals[pt2+1] = 0.0;          // real
	    tvals[pt2] = tmp; 
	    tvals[pt1 + 1] = -tmp;        // image (form 1)
	  }
	  else {
	    tvals[pt1 + 1] = tvals[pt2]     = 0.0;   // image 
	    tvals[pt1] = tmp;   
	    tvals[pt2 + 1] = tmp;   // real (form 1)
	  }
	}
      }
      assert(cj==ndfj_tot);
    }
  }
  assert(pt1 <= 4096); 
  ierr = PromMatrix::SetValues(ndfi_tot,pgidi,ndfj_tot,pgidj,tvals,add_t);
  CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::GetNodeScalarRow_PETSc" 
/* PromMatrix::GetNodeScalarRow_PETSc *****************************************
 *
 *  - Get rows (of node) of PETSc Mat.
 *
 *  INPUT:
 *    - ceq: local
 *    - len: buffer length(s)
 *    - ncols: (out)
 *    - Brp: (out)
 *    - Bap: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromMatrix::GetNodeScalarRow_PETSc( const int ceq, const int len, int &ncols,
					int Brp[], double *Bap[] /*=NULL*/ )const
{
  Mat matin = mat_;
  double      *aa;
  int         *rp,my0,np,ierr;
  int         *ai,*ailen,*aj,nn,kk,nrow, brow = ceq; assert(brow>=0);
  Mat_SeqAIJ  *mat;   
  ierr = MatGetOwnershipRange( mat_, &my0, &kk ); CHKERRQ(ierr);
  MPI_Comm_size(mat_->comm,&np); 

  // straight copy 
  if( np > 1 ) {
    Mat_MPIAIJ *aij =(Mat_MPIAIJ*)matin->data;  
    mat = (Mat_SeqAIJ*)aij->B->data; 
    ai = mat->i; aj = mat->j;  ailen = mat->ilen;  
    nrow = ailen[brow]; 
    for( nn = kk = 0, aa = mat->a + ai[brow], rp = aj + ai[brow]; 
	 kk != nrow; kk++, aa += 1 ){
      int lid = *rp++, gid = aij->garray[lid];
      if( Bap != NULL ) Bap[nn] = aa; 
      Brp[nn] = gid; 
      nn++;
    }
    mat = (Mat_SeqAIJ*)aij->A->data; 
  }
  else {
    nn = 0;
    mat = (Mat_SeqAIJ*)matin->data; 
  }

  // local
  ai = mat->i; aj = mat->j; ailen = mat->ilen;  
  nrow = ailen[brow]; 
  if( nn+nrow > len ) {
    SETERRQ2(1,"ERROR: ncols(%d) > len(%d)",nn+nrow,len);
  }
  for( kk = 0, aa = mat->a + ai[brow], rp = aj + ai[brow]; 
       kk != nrow; kk++, aa += 1 ){
    int lid = *rp++, gid = lid + my0;
    if( Bap ) Bap[nn] = aa; 
    Brp[nn] = gid;
    nn++;
  }
  ncols = nn; // out

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::GetLocal_constBlk_NodeRow" 
/* PromMatrix::GetLocal_constBlk_NodeRow *************************************
 *
 */
int PromMatrix::GetLocal_constBlk_NodeRow( const int MyRow, const int len, 
					   int &ncols, int adjacs[],
					   double **Ai ) const
{
  int ierr, my0, tt;
  MatType type;
  ierr = MatGetBlockSize( mat_, &tt );               CHKERRQ( ierr );
  const int bs = tt, np = map_.comm_.np(), bs2 = bs * bs;
  
  if( bs == 1 ) {
    ierr = MatGetType( mat_, &type );  CHKERRQ( ierr );
    if( strcmp(type, MATMPIAIJ) == 0 || strcmp(type, MATSEQAIJ) == 0) {
      return GetNodeScalarRow_PETSc(MyRow, len, ncols, adjacs, Ai);
    }
    else SETERRQ1(1,"Wrong type: %s",type);
  }
  // straight copy
  ierr = MatGetOwnershipRange( mat_, &my0, &tt ); CHKERRQ(ierr);
  assert(my0%bs==0); my0 = my0/bs;
  if( MyRow < 0 ) SETERRQ1(1,"ERROR: myRow=%d",MyRow);

  double      *aa; 
  int         *ai,*ailen,*aj,kk,nn,*rp,nc;  
  Mat_SeqBAIJ *mat;
  if( np > 1 ) { // set 'nn' and 'mat'
#if defined(PETSC_USE_BOPT_g)
    if(strcmp(type, MATMPIBAIJ) != 0)SETERRQ1(1,"Wrong type: %s",type);
#endif
    // off local 
    Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)mat_->data; // upper left matix  
    mat = (Mat_SeqBAIJ*)baij->B->data;
    ai = mat->i; ailen = mat->ilen;  nc = ailen[MyRow]; aj = mat->j;
    assert( baij->garray != NULL );
    for( nn = kk = 0, aa = mat->a + bs2 * ai[MyRow], rp = aj + ai[MyRow];    
	 kk != nc ; kk++, aa += bs2 ){
      int lid=*rp++,gid=baij->garray[lid];
      if( Ai ) Ai[nn] = aa;
      adjacs[nn] = gid;
      nn++;
    }
    mat = (Mat_SeqBAIJ*)baij->A->data; // main matrix
  }
  else{ // one proc
    ierr = MatGetType( mat_, &type );  CHKERRQ( ierr );
    nn = 0; 
    if (strcmp(type, MATSEQBAIJ) == 0) {
      mat = (Mat_SeqBAIJ*) mat_->data;      // Handle SeqBAIJ
    }
    else if( strcmp(type, MATMPIBAIJ) == 0 ){
      Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)mat_->data; // upper left matix  
      mat = (Mat_SeqBAIJ*) baij->A->data; // Handle MPIBAIJ on one proc?
    }
    else SETERRQ1(1,"Wrong type: %s",type);
  }
  
  // local
  ai = mat->i; ailen = mat->ilen;  nc = ailen[MyRow]; aj = mat->j;
  if( nn+nc > len ) {
    SETERRQ2(1,"ERROR: ncols(%d) > PROM_MAXNIEGH(%d)",nn+nc,len);
  }
  else if(nc==0)PetscPrintf(PETSC_COMM_SELF,
			    "[%d]%s Warning zero local row %d\n",
			    map_.comm_.mype(),__FUNCT__,MyRow);
  for( kk = 0, aa = mat->a + bs2 * ai[MyRow], rp = aj + ai[MyRow] ; 
       kk != nc ; kk++, aa += bs2 ){
    int lid = *rp++,gid=lid+my0;
    if( Ai ) Ai[nn] = aa; 
    adjacs[nn] = gid;
    nn++;
  }
  ncols = nn; // ret
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::ConvertToSymm" 
/* PromMatrix::ConvertToSymm *************************************
 *
 */
int PromMatrix::ConvertToSymm()
{
  if( isSymmetricMatrix() ) return 0;
  assert(this!=NULL && mat_!=NULL);
  //
  int ierr,ncols,ii,adjacs[PROM_MAXNIEGH], igid, eq0=map_.my0Eq(), neq = map_.nlocalEq();
  // create
  ierr = MatGetBlockSize( mat_, &ii );               CHKERRQ( ierr );
  const int bs=ii, myb0=eq0/bs, nbloc=neq/bs, nextb0 = myb0 + nbloc;
  int *nnz=NULL,*onnz=NULL;  assert(neq%bs==0);

  if( nbloc > 0 ) {
    // get estimate of nnz
    ierr = PetscMalloc( sizeof(int)*(nbloc), &nnz ); CHKERRQ( ierr );
    ierr = PetscMalloc( sizeof(int)*(nbloc), &onnz ); CHKERRQ( ierr );
    for( ii = 0 ; ii < nbloc ; ii++ ) onnz[ii] = nnz[ii] = 0;
  }
  for( ii = 0, igid = myb0 ; ii < nbloc ; ii++, igid++ ) {
    ierr = GetLocal_constBlk_NodeRow( ii, PROM_MAXNIEGH, ncols, adjacs ); CHKERRQ(ierr);
    for( int jj = 0 ; jj < ncols ; jj++ ) {
      int jgid = adjacs[jj];
      if( jgid >= nextb0 ) onnz[ii]++;
      else if( jgid >= igid ) nnz[ii]++;
    }
    assert( nnz[ii] > 0);
  }
  Mat symm; ii = bs*nbloc; ncols = map_.nglobalEq();
  ierr = MatCreateMPISBAIJ( mat_->comm, bs, ii, ii, ncols, ncols, 0, nnz, 
			    0, onnz, &symm );
  CHKERRQ( ierr );
  ierr = MatZeroEntries( symm );   CHKERRQ(ierr);  
  if( nnz != NULL ) PetscFree(nnz);
  if( onnz != NULL ) PetscFree(onnz);
  ierr = MatSetOption( symm, MAT_NEW_NONZERO_ALLOCATION_ERR ); CHKERRQ(ierr);
  ierr = MatSetOption( symm, MAT_IGNORE_OFF_PROC_ENTRIES); CHKERRQ(ierr); // not needed
  //

  double *Ai[PROM_MAXNIEGH],buf[4*PROM_MAXNDF*PROM_MAXNDF]; 
  for( ii = 0, igid = myb0 ; ii < nbloc ; ii++, igid++ ) {
    ierr = GetLocal_constBlk_NodeRow( ii, PROM_MAXNIEGH, ncols, adjacs, Ai);
    CHKERRQ(ierr);
    for( int jj = 0 ; jj < ncols ; jj++ ) {
      int jgid = adjacs[jj];
      if( jgid >= igid ){
	const double *aijT = Ai[jj]; // transpose
	for(int ki = 0 ; ki < bs ; ki++ ) {
	  for(int kj = 0 ; kj < bs ; kj++) buf[ki*bs+kj] = aijT[kj*bs+ki]; 
	}
	ierr = MatSetValuesBlocked(symm, 1, &igid, 1, &jgid,buf,INSERT_VALUES);
	CHKERRQ(ierr);
      }
    }
  }

  //ierr = MatDestroy( mat_ ); CHKERRQ(ierr); mat_ = NULL;
  assert(full_mat_==NULL); full_mat_ = mat_;
  //
  ierr = MatAssemblyBegin( symm, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
  ierr = MatAssemblyEnd( symm, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
  mat_ = symm;

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::ConvertFromSymm" 
/* PromMatrix::ConvertFromSymm *************************************
 *
 */
int PromMatrix::ConvertFromSymm()
{ 
  if( !isSymmetricMatrix() ) return 0;
  //else if( full_mat_ == NULL ){ SETERRQ(1,"full_mat==NULL"); }
  else {
    int ierr = MatDestroy( mat_ ); CHKERRQ(ierr);
    mat_ = full_mat_; full_mat_ = NULL; 
    ierr = MatZeroEntries( mat_ ); CHKERRQ(ierr);
    return 0;
  }
  /* causes some bug!!!
  assert(this!=NULL && mat_!=NULL); const int tag = map_.comm_.getNewTag();
  int ierr,ncols,ii,adjacs[PROM_MAXNIEGH], igid, proc;
  PromTable<PromList<int>*> proc_list(57); PromList<int> *list;
  PromList<MPI_Request*> smessages; MPI_Status status; 
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; const int np=map_.comm_.np();
  int  reqsz = sizeof(MPI_Request)/sizeof(int),mype=map_.comm_.mype();
  MPI_Comm wcomm = map_.comm_.MPIComm();

  // create
  int *nnz=NULL,*onnz=NULL;         MatGetBlockSize( mat_, &ii ); 
  const int nloc=map_.nlocal(),bs=ii,my0=map_.my0Nd(),next0=my0+nloc;
  // get estimate of nnz
  ierr = PetscMalloc( sizeof(int)*(nloc+1), &nnz ); CHKERRQ( ierr );
  ierr = PetscMalloc( sizeof(int)*(nloc+1), &onnz ); CHKERRQ( ierr );
  for( ii = 0 ; ii < nloc ; ii++ ) onnz[ii] = nnz[ii] = 0;
  onnz[ii] = nnz[ii] = -9;
  for( ii = 0, igid = my0 ; ii < nloc ; ii++, igid++ ) {
    ierr = GetLocalNodeRow(ii, PROM_MAXNIEGH, ncols, adjacs);CHKERRQ(ierr);
    for( int jj = 0 ; jj < ncols ; jj++ ) {
      int jgid = adjacs[jj], kk = jgid - my0;
      if( kk >= 0 && kk < nloc ){ 
	assert(kk>=ii);
	nnz[ii]++; 
	if( ii != kk ) nnz[kk]++; 
      }
      else{
	onnz[ii]++; 
	ierr = PromGrid::FindProc( map_.proc_gnode(), np, jgid, proc ); CHKERRQ(ierr);
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<int>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( jgid + 1 );
	assert(proc!=mype);
      }
    }
    assert( nnz[ii] > 0);
  }
  // send(tag) *< gid >
  for( ii = 0 ; ii < np ; ii++ ) sbuf[ii] = 0;  
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    int tt = list->getCount(); 
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) *pb++ = list->removeHead() - 1;
    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    // send(tag) *< gid >
    MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
    delete list;
    sbuf[proc] = tt; assert(tt>0);
  }
  // recive requests -- global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, wcomm );
  for( proc = 0 ; proc < np ; proc++ ) {
    if( rbuf[proc] != 0 ) {
      // recv(tag)
      int tt = rbuf[proc], *buff, *pb; 
      ierr = PetscMalloc(tt*sizeof(int),&buff );  CHKERRQ(ierr);
      pb = buff;
      // recv(tag) *< gid >
      MPI_Recv( buff, tt, MPI_INT, proc, tag, wcomm, &status );
      while( pb-buff < tt ) {
	igid = *pb++; assert(igid >= my0 && igid < next0);
	onnz[igid - my0]++;
      }
      assert(pb-buff == tt);       PetscFree(buff);
    }
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &ii, &status );  assert(!ii);
  //
  Mat nsymm; ii = map_.nlocalEq(); ncols = map_.nglobalEq();
  ierr = MatCreateMPIBAIJ( mat_->comm, bs, ii, ii, ncols, ncols, 0, nnz, 
			   0, onnz, &nsymm );
  CHKERRQ( ierr );
  ierr = MatSetOption( nsymm, MAT_NEW_NONZERO_ALLOCATION_ERR ); CHKERRQ(ierr);
  ierr = MatZeroEntries( nsymm );   CHKERRQ(ierr);
  assert(onnz[nloc] == nnz[nloc] && nnz[nloc] == -9);
  PetscFree(nnz); PetscFree(onnz);
  //
  ierr = MatDestroy( mat_ ); CHKERRQ(ierr);
  mat_ = nsymm;

  return 0;*/
}

#undef __FUNCT__  
#define __FUNCT__ "PromMatrix::GetSymmetricLocal_constBlk_NodeRow" 
/* PromMatrix::GetSymmetricLocal_constBlk_NodeRow *************************************
 *
 */
int PromMatrix::GetSymmetricLocal_constBlk_NodeRow( const int MyRow, const int len, 
					  int &ncols, int adjacs[],
					  double **Ai ) const
{
  assert(0); // not used
  int         *ai,*ailen,*aj,kk,nn,*rp,nrow,ierr;
  ierr = MatGetBlockSize( mat_, &kk ); CHKERRQ(ierr);
  const int bs=kk,myb0=map_.my0Eq()/bs,np=map_.comm_.np(), bs2 = bs*bs; 
  Mat AA = mat_;
  Mat_MPISBAIJ *baij = (Mat_MPISBAIJ*)AA->data; // upper left matix
  double      *aa; 
  Mat_SeqSBAIJ *mat;
  
  // straight copy
  if( np > 1 ) {
    // off local 
    Mat_SeqBAIJ *matb = (Mat_SeqBAIJ*)baij->B->data;
    ai = matb->i; ailen = matb->ilen;  nrow = ailen[MyRow]; aj = matb->j;
    assert( baij->garray != NULL );
    for( nn = kk = 0, aa = matb->a + bs2 * ai[MyRow], rp = aj + ai[MyRow];    
	 kk != nrow ; kk++, aa += bs2 ){
      int lid=*rp++,gid=baij->garray[lid];
      if( Ai ) Ai[nn] = aa;
      adjacs[nn] = gid;
      nn++;
    }
    mat = (Mat_SeqSBAIJ*)baij->A->data; // main matrix
  }
  else{ 
    mat = (Mat_SeqSBAIJ*)AA->data; nn = 0; 
  }
  //assert(bs==mat->bs);
  // local
  ai = mat->i; ailen = mat->ilen;  nrow = ailen[MyRow]; aj = mat->j;
  assert(nrow>=0);
  if( nn+nrow > len ) {
    SETERRQ2(1,"ERROR: ncols(%d) > PROM_MAXNIEGH(%d)",nn+nrow,len);
  }
  else if(nrow==0)PetscPrintf(PETSC_COMM_SELF,
			      "[%d]%s Warning zero local row %d\n",
			      map_.comm_.mype(),__FUNCT__,MyRow);
  for( kk = 0, aa = mat->a + bs2 * ai[MyRow], rp = aj + ai[MyRow] ; 
       kk != nrow ; kk++, aa += bs2 ){
    int lid = *rp++,gid=lid+myb0;
    if( Ai ) Ai[nn] = aa; 
    adjacs[nn] = gid;
    nn++;
  }
  ncols = nn; // ret

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromScatter::Scatter" 
/* PromScatter::Scatter *****************************************
 *
 *  INPUT:
 *    - X: local
 *    - Y: global
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromScatter::Scatter( const PromSeqVector *X, PromVector_base *Y )
{
  assert(this!=NULL);
  PromComplexVector *cY = dynamic_cast<PromComplexVector*>(Y); 
  assert( cY == NULL );
  PromVector *YY = dynamic_cast<PromVector*>(Y);
  int ierr = scatter_private( SCATTER_FORWARD, X->svec_, YY->vec_ );
  CHKERRQ(ierr);
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromScatter::Scatter" 
/* PromScatter::Scatter *****************************************
 *
 *  INPUT:
 *    - X: local
 *    - Y: global
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromScatter::Gather( PromSeqVector *X, const PromVector_base *Y )
{
  assert(this!=NULL); int ierr;
  const PromComplexVector *cY = dynamic_cast<const PromComplexVector*>(Y); 
  if( cY != NULL ) {
    ierr = Gather_r( X, Y );  CHKERRQ(ierr); assert(0);
  }
  else {
    const PromVector *YY = dynamic_cast<const PromVector*>(Y);
    ierr = scatter_private(SCATTER_REVERSE, X->svec_, YY->vec_); CHKERRQ(ierr);
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromScatter::Gather_i" 
/* PromScatter::Gather_i *****************************************
 *
 *  INPUT:
 *    - X: local
 *    - Y: global
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromScatter::Gather_i( PromSeqVector *X, const PromVector_base *Y)
{
  assert(this!=NULL);
  int ierr, ii, sz, *ids, *ids_i;
  const PromVector *YY = dynamic_cast<const PromVector*>(Y);
  if( IS_i_ == NULL ) {
    ierr = ISGetIndices( ISy_, &ids ); CHKERRQ(ierr);
    ierr = ISGetLocalSize( ISy_, &sz ); CHKERRQ(ierr);
    ierr = PetscMalloc( (sz+1)*sizeof(int), &ids_i ); CHKERRQ(ierr);
    for(ii=0;ii<sz;ii++) ids_i[ii] = 2*ids[ii]+1;
    ierr = ISRestoreIndices( ISy_, &ids ); CHKERRQ(ierr);
    PromISCreateGeneral( comm_(), sz, ids_i, &IS_i_ );       
    ierr = PetscFree( ids_i ); CHKERRQ(ierr);
  }
  if( scat_i_ == 0 ) {
    ierr = VecScatterCreate( X->svec_, ISx_, YY->vec_, IS_i_, &scat_i_); 
    CHKERRQ(ierr);
  }
  ierr = VecScatterBegin(scat_i_, YY->vec_, X->svec_, INSERT_VALUES, SCATTER_REVERSE);
  CHKERRQ(ierr);
  ierr = VecScatterEnd(scat_i_, YY->vec_, X->svec_, INSERT_VALUES, SCATTER_REVERSE); 
  CHKERRQ(ierr);
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromScatter::Gather_r" 
/* PromScatter::Gather_r *****************************************
 *
 *  INPUT:
 *    - X: local
 *    - Y: global
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *      - PETSc error code
 */
int PromScatter::Gather_r( PromSeqVector *X, const PromVector_base *Y)
{
  assert(this!=NULL);
  int ierr, ii, sz, *ids, *ids_r;
  const PromVector *YY = dynamic_cast<const PromVector*>(Y);
  if( IS_r_ == NULL ) {
    ierr = ISGetIndices( ISy_, &ids ); CHKERRQ(ierr);
    ierr = ISGetLocalSize( ISy_, &sz ); CHKERRQ(ierr);
    ierr = PetscMalloc( (sz+1)*sizeof(int), &ids_r ); CHKERRQ(ierr);
    for(ii=0;ii<sz;ii++) ids_r[ii] = 2*ids[ii];
    ierr = ISRestoreIndices( ISy_, &ids ); CHKERRQ(ierr);
    PromISCreateGeneral( comm_(), sz, ids_r, &IS_r_ );
    ierr = PetscFree( ids_r ); CHKERRQ(ierr);
  }
  if( scat_r_ == NULL ) {
    ierr = VecScatterCreate( X->svec_, ISx_, YY->vec_, IS_r_, &scat_r_); 
    CHKERRQ(ierr);
  }
  ierr = VecScatterBegin(scat_r_, YY->vec_, X->svec_, INSERT_VALUES, SCATTER_REVERSE);
  CHKERRQ(ierr);
  ierr = VecScatterEnd(scat_r_, YY->vec_, X->svec_, INSERT_VALUES, SCATTER_REVERSE); 
  CHKERRQ(ierr);
  
  return 0;
}

#else  // PETRA

//! PromIS methods for index sets for PC blocks
int PromISGetSize( PromIS is, int *out )
{
  *out = is->NumMyElements();
  return 0;
}
int PromISGetIndices( PromIS is, int *out[] )
{
  int ierr, sz = is->NumMyElements();
  ierr = PetscMalloc( sz*sizeof(int), out ); CHKERRQ(ierr); 
  is->MyGlobalElements( *out ); 
  return 0;
}
int PromISRestoreIndices( PromIS, int *out[] )
{
  int ierr = PetscFree( *out ); CHKERRQ(ierr);
  *out = NULL;
  return 0;
}
int PromISCreateGeneral( MPI_Comm, int lsz, int *gids,
			 PromIS *out )
{ 
  Epetra_Map *map = new Epetra_Map( -1, lsz, gids, 0, Comm ); 
  *out = map;
  return 0;
}
int PromISDestroy( PromIS is )
{
  delete is;
  return 0;
}

#endif

#if !defined(PROM_HAVE_PETSC)

#undef __FUNCT__  
#define __FUNCT__ "PetscPrintf" 
/*************************************************************************
 * Petsc printf
 *************************************************************************/
int PetscPrintf(MPI_Comm comm,const char format[],...)
{
  int       rank,ierr,len;
  char      *nformat,*sub1,*sub2;
  PetscReal value;

  PetscFunctionBegin;
  if( comm == MPI_COMM_NULL ) comm = PETSC_COMM_WORLD;
  ierr = MPI_Comm_rank(comm,&rank);CHKERRQ(ierr);
  if (rank == 0) {
    va_list Argp;
    va_start(Argp,format);
    
    ierr = PetscStrstr(format,"%A",&sub1);CHKERRQ(ierr);
    if (sub1) {
      ierr = PetscStrstr(format,"%",&sub2);CHKERRQ(ierr);
      if(sub1 != sub2) SETERRQ(1,"%A format must be first in format string");
      ierr    = PetscStrlen(format,&len);CHKERRQ(ierr);
      ierr =  PetscMalloc((len+16)*sizeof(char),&nformat);CHKERRQ(ierr);
      ierr    = PetscStrcpy(nformat,format);CHKERRQ(ierr);
      ierr    = PetscStrstr(nformat,"%",&sub2);CHKERRQ(ierr);
      sub2[0] = 0;
      value   = (double)va_arg(Argp,double);
      if (PetscAbsDouble(value) < 1.e-12) {
        ierr    = PetscStrcat(nformat,"< 1.e-12");CHKERRQ(ierr);
      } else {
        ierr    = PetscStrcat(nformat,"%g");CHKERRQ(ierr);
        va_end(Argp);
        va_start(Argp,format);
      }
      ierr    = PetscStrcat(nformat,sub1+2);CHKERRQ(ierr);
    } else {
      nformat = (char*)format;
    }
#if defined(PETSC_HAVE_VPRINTF_CHAR)
    vfprintf(stdout,nformat,(char *)Argp);
#else
    vfprintf(stdout,nformat,Argp);
#endif
    fflush(stdout);
    if (petsc_history) {
#if defined(PETSC_HAVE_VPRINTF_CHAR)
      vfprintf(petsc_history,nformat,(char *)Argp);
#else
      vfprintf(petsc_history,nformat,Argp);
#endif
      fflush(petsc_history);
    }
    va_end(Argp);
    if (sub1) {ierr = PetscFree(nformat);CHKERRQ(ierr);}
  }
  PetscFunctionReturn(0);
}
#endif
